#! /usr/bin/python3.6
import logging
import statistics

from collections import defaultdict
from typing import Dict, List
from autobridge.Device.DeviceManager import *
from autobridge.Opt.Common import RESOURCE_TYPES
from autobridge.Opt.FloorplanLegalize import AutoLegalizer
from autobridge.Opt.DataflowGraph import *
from autobridge.Opt.Slot import Slot
from autobridge.Opt.SlotManager import SlotManager, Dir
from autobridge.HLSParser.vivado_hls.HLSProjectManager import HLSProjectManager
from mip import *

_logger = logging.getLogger().getChild(__name__)


class Floorplanner:

  def __init__(
      self,
      graph : DataflowGraph,
      user_constraint_s2v : Dict,
      slot_manager : SlotManager,
      total_usage : dict,
      board,
      user_max_usage_ratio = 0.7,
      max_search_time=600,
      grouping_hints=[],
      grouping_constraints=[]):
    self.board = board
    self.graph = graph
    self.user_constraint_s2v = user_constraint_s2v
    for slot, v_list in user_constraint_s2v.items():
      for v in v_list:
        _logger.info(f'user constraints: {v.name} -> {slot.getRTLModuleName()}')

    self.slot_manager = slot_manager
    self.total_usage = total_usage
    self.max_search_time = max_search_time

    # hints are optional, constraints are mandatory
    self.grouping_hints = grouping_hints # list of list
    self.grouping_constraints: List[List[str]] = grouping_constraints
    if grouping_constraints:
      for group in grouping_constraints:
        _logger.info(f'initial grouping constraints: ' + ' & '.join(group))
    else:
      _logger.info(f'no initial grouping constraints')

    self.s2v = {}
    self.v2s = {}
    self.s2e = {}

    # incrementally adjust "delta" by "step" to calibrate the resource usage limit
    self.delta = 0
    self.step = 0.03
    self.MAX_USAGE_ALLOWED = 0.85 # cut the process if surpass this value
    self.max_usage_ratio = self.__getResourceUsageLimit(user_max_usage_ratio)

    self.__checker()

  def __checker(self):
    for v_group in self.user_constraint_s2v.values():
      for v in v_group:
        assert v in self.graph.getAllVertices(), f'{v.name} is not a valid RTL module'

  def __getResourceUsageLimit(self, user_max_usage_ratio):
    total_avail = self.board.TOTAL_AREA

    for item in RESOURCE_TYPES:
      usage = self.total_usage[item] / total_avail[item] + 0.05
      ratio = max(usage, user_max_usage_ratio)

    _logger.info(f'Maximum resource usage ratio set as: {ratio}')
    return ratio

  def __initSlotToEdges(self):
    self.s2e = {}
    for s, v_group in self.s2v.items():
      assert v_group, f'incorrect empty slot: {s.getName()}'

      # add the edge if the src and dst are both in this slot
      intra_edges, inter_edges = self.getIntraAndInterEdges(v_group)
      self.s2e[s] = intra_edges

      # for the FIFO connecting two different slots, it should be assigned to the destination side
      v_set = set(v_group)
      for e in inter_edges:
        if e.dst in v_set:
          self.s2e[s].append(e)
          _logger.debug(f'{e.name} is assigned with {e.dst.name}')

  # map all vertices to the initial slot (the whole device)
  def __getInitialSlotToVerticesMapping(self):
    init_slot = self.slot_manager.getInitialSlot()
    init_s2v = {init_slot : self.graph.getAllVertices()}
    init_v2s = {v : init_slot for v in self.graph.getAllVertices()}
    return init_s2v, init_v2s

  def __addAreaConstraints(self, m, curr_s2v, v2var, dir, delta=0):
    for s, v_group in curr_s2v.items():
      bottom_or_left, up_or_right = self.slot_manager.partitionSlotByHalf(s, dir)
      assert up_or_right.up_right_x >= bottom_or_left.down_left_x

      for r in RESOURCE_TYPES:
        v_var_list = [v2var[v] for v in v_group]
        area_list = [v.area[r] for v in v_group]
        I = range(len(v_group))

        # for the up/right child slot (if mod_x is assigned 1)
        m += xsum( v_var_list[i] * area_list[i] for i in I ) <= up_or_right.getArea()[r] * (self.max_usage_ratio + delta)

        # for the down/left child slot (if mod_x is assigned 0)
        m += xsum( (1-v_var_list[i]) * area_list[i] for i in I ) <= bottom_or_left.getArea()[r] * (self.max_usage_ratio + delta)

  def __addUserConstraints(self, m, curr_v2s, v2var, dir, check_user_constraints = True):
    for expect_slot, v_group in self.user_constraint_s2v.items():
      for v in v_group:
        # this corner case is for separate partition of small slots
        if not check_user_constraints:
          if v not in curr_v2s:
            continue

        assert v in curr_v2s, f'ERROR: user has forced the location of a non-existing module {v.name}'

        curr_slot = curr_v2s[v]
        bottom_or_left, up_or_right = self.slot_manager.partitionSlotByHalf(curr_slot, dir)
        if bottom_or_left.containsChildSlot(expect_slot):
          _logger.debug(f'[user constraint] {v.name} assigned to bottom/left')
          m += v2var[v] == 0
        elif up_or_right.containsChildSlot(expect_slot):
          _logger.debug(f'[user constraint] {v.name} assigned to up/right')
          m += v2var[v] == 1
        else:
          _logger.warning(f'Potential wrong constraints from user: {v.name} -> {expect_slot.getName()}')
          exit(1)

  # specify which modules must be assigned to the same slot
  # note that the key of curr_v2s is Vertex instead of name string
  def __addGroupingConstraints(self, m, curr_v2s, v2var, enable_grouping_hints):
    _logger.info(f'apply grouping constraints')
    v_name_to_v = {v.name : v for v in curr_v2s.keys()}

    if enable_grouping_hints:
      target = self.grouping_hints + self.grouping_constraints
    else:
      target = self.grouping_constraints

    for grouping in target:
      # assert grouping[0] in v_name_to_v, f'unknown vertex: {grouping[0]}'
      for i in range(1, len(grouping)):
        # assert grouping[i] in v_name_to_v, f'unknown vertex: {grouping[i]}'

        if grouping[0] in v_name_to_v and grouping[i] in v_name_to_v:
          _logger.info(f'[grouping] {grouping[0]} is grouped with {grouping[i]}')
          m += v2var[v_name_to_v[grouping[0]]] == v2var[v_name_to_v[grouping[i]]]

  def __addOptGoal(self, m, curr_v2s, external_v2s, v2var, dir):
    def getVertexPosInChildSlot(v : Vertex):
      def getChildSlotPositionX(v):
        if v in external_v2s:
          return external_v2s[v].getPositionX() # const
        else:
          return curr_v2s[v].getQuarterPositionX() + v2var[v] * curr_v2s[v].getHalfLenX() # expr

      def getChildSlotPositionY(v):
        if v in external_v2s:
          return external_v2s[v].getPositionY() # const
        else:
          return curr_v2s[v].getQuarterPositionY() + v2var[v] * curr_v2s[v].getHalfLenY() # expr

      if dir == Dir.vertical:
        return getChildSlotPositionX(v)
      elif dir == Dir.horizontal:
        return getChildSlotPositionY(v)
      else:
        assert False

    # get all involved edges. Differentiate internal edges and boundary edges
    intra_edges, interface_edges = self.getIntraAndInterEdges(curr_v2s.keys())

    # cost function.
    edge_costs = [m.add_var(var_type=INTEGER, name=f'intra_{e.name}') for e in intra_edges] \
        + [m.add_var(var_type=INTEGER, name=f'inte_{e.name}') for e in interface_edges]

    all_edges = intra_edges + interface_edges
    for e_cost, e in zip(edge_costs, all_edges):
      m += e_cost >= getVertexPosInChildSlot(e.src) - getVertexPosInChildSlot(e.dst)
      m += e_cost >= getVertexPosInChildSlot(e.dst) - getVertexPosInChildSlot(e.src)

    m.objective = minimize(xsum(edge_costs[i] * edge.width for i, edge in enumerate(all_edges) ) )

  def __getPartitionResult(self, num_solutions, curr_s2v, v2var, dir):
    def getIthSolution(i, temp_slots):
      # create new mapping
      next_s2v = {}
      next_v2s = {}

      for s, v_group in curr_s2v.items():
        bottom_or_left, up_or_right = self.slot_manager.partitionSlotByHalf(s, dir)
        temp_slots += [bottom_or_left, up_or_right]

        next_s2v[bottom_or_left] = []
        next_s2v[up_or_right] = []
        for v in v_group:
          # if v is assigned to 0-half in the i-th solution
          if round(v2var[v].xi(i)) == 0:
            next_s2v[bottom_or_left].append(v)
            next_v2s[v] = bottom_or_left

          # if v is assigned to 1-half in the i-th solution
          elif round(v2var[v].xi(i)) == 1:
            next_s2v[up_or_right].append(v)
            next_v2s[v] = up_or_right
          else:
            assert False, v2var[v].xi(i)

          # sometimes the result is not strictly integer?
          assert abs(round(v2var[v].xi(i)) - v2var[v].xi(i)) < 0.0001, v2var[v].xi(i)

        # if no Vertex is assigned to a Slot, remove that Slot
        if not next_s2v[bottom_or_left]:
          next_s2v.pop(bottom_or_left)
          self.slot_manager.removeSlotNonBlocking(bottom_or_left.getName())
        if not next_s2v[up_or_right]:
          next_s2v.pop(up_or_right)
          self.slot_manager.removeSlotNonBlocking(up_or_right.getName())

      return next_s2v, next_v2s

    _logger.info(f'there are {num_solutions} solutions available')

    best_next_s2v = {}
    best_next_v2s = {}
    best_var = float('inf')

    # track all temp slots and delete the empty ones at the end
    temp_slots = []

    # choose the most balanced solution
    for i in range(num_solutions):
      next_s2v, next_v2s = getIthSolution(i, temp_slots)

      # get the average variance of each resource
      var = 0
      for r in RESOURCE_TYPES:
        # percentage of resource r
        s2area = {slot : sum(v.area[r] for v in v_list) / slot.getArea()[r] \
                  for slot, v_list in next_s2v.items() }

        # corner case: if next_s2v only has one slot
        expect_slot_num = len(curr_s2v) * 2
        empty_slot_num = expect_slot_num - len(s2area)
        slots_usage_list = list(s2area.values()) + [0] * empty_slot_num

        var += statistics.variance(slots_usage_list)


      _logger.info(f'the {i}-th solution has variance: {var}')

      # track the best solution
      if var < best_var:
        best_var = var
        best_next_s2v = next_s2v
        best_next_v2s = next_v2s

    # remove unused temp slots
    for temp in temp_slots:
      if temp not in best_next_s2v:
        self.slot_manager.removeSlotNonBlocking(temp.getName())

    return best_next_s2v, best_next_v2s

  def __createILPVariables(self, m, curr_v2s):
    v2var = {} # str -> [mip_var]
    for v in curr_v2s.keys():
      v2var[v] = m.add_var(var_type=BINARY, name=f'{v.name}_x')

    return v2var

  # could only be invoked at the beginning when there is only one slot
  def eightWayPartition(self):
    _logger.info('Start 8-way partitioning routine')

    curr_s2v, curr_v2s = self.__getInitialSlotToVerticesMapping()

    m = Model()
    if not _logger.isEnabledFor(logging.DEBUG):
      m.verbose = 0

    # three variables could determine the location of a module
    # y = y1 *2 + y2  (four slots)
    # x = x           (each SLR is divided by half)
    v2var_x = {}
    v2var_y1 = {}
    v2var_y2 = {}
    for v in curr_v2s.keys():
      v2var_x[v] = m.add_var(var_type=BINARY, name=f'{v.name}_x')
      v2var_y1[v] = m.add_var(var_type=BINARY, name=f'{v.name}_y1')
      v2var_y2[v] = m.add_var(var_type=BINARY, name=f'{v.name}_y2')

    # get the target slots
    # for U280, slot_11x will have empty area
    init_slot = self.slot_manager.getInitialSlot()
    slot_0, slot_1 = self.slot_manager.getBottomAndUpSplit(init_slot)

    slot_00, slot_01 = self.slot_manager.getBottomAndUpSplit(slot_0)
    slot_10, slot_11 = self.slot_manager.getBottomAndUpSplit(slot_1)

    slot_000, slot_001 = self.slot_manager.getLeftAndRightSplit(slot_00)
    slot_010, slot_011 = self.slot_manager.getLeftAndRightSplit(slot_01)
    slot_100, slot_101 = self.slot_manager.getLeftAndRightSplit(slot_10)
    slot_110, slot_111 = self.slot_manager.getLeftAndRightSplit(slot_11)

    # must not change order!
    slot_group = [slot_000, slot_001, \
                  slot_010, slot_011, \
                  slot_100, slot_101, \
                  slot_110, slot_111 ]
    # note that slot_idx is different from slot position
    slot_idx = lambda y1, y2, x : y1 * 4 + y2 * 2 + x

    # area constraint
    for r in RESOURCE_TYPES:
      choose = lambda x, num: x if num == 1 else (1-x)

      for y1 in range(2):
        for y2 in range(2):
          for x in range(2):
            # convert logic AND to linear constraints
            # prods[v] = choose_y1 AND choose_y2 AND choose_x
            prods = { v : m.add_var(var_type=BINARY, name=f'{v.name}_choose{y1}{y2}{x}') for v in curr_v2s.keys() }
            for v in curr_v2s.keys():
              m +=  choose(v2var_y1[v], y1) + choose(v2var_y2[v], y2) + \
                    choose(v2var_x[v], x) - 3 * prods[v] >= 0
              m +=  choose(v2var_y1[v], y1) + choose(v2var_y2[v], y2) + \
                    choose(v2var_x[v], x) - 3 * prods[v] <= 2

            m += xsum(  prods[v] * v.getVertexAndInboundFIFOArea()[r] for v in curr_v2s.keys() ) \
                        <= slot_group[slot_idx(y1, y2, x)].getArea()[r] * self.max_usage_ratio

    # user constraint
    for expect_slot, v_group in self.user_constraint_s2v.items():
      for v in v_group:
        assert v in curr_v2s, f'ERROR: user has forced the location of a non-existing module {v.name}'

        for y1 in range(2):
          for y2 in range(2):
            for x in range(2):
              if slot_group[slot_idx(y1, y2, x)].containsChildSlot(expect_slot):
                m += v2var_y1[v] == y1
                m += v2var_y2[v] == y2
                m += v2var_x[v] == x

    # grouping constraints
    for v2var in [v2var_x, v2var_y1, v2var_y2]:
      self.__addGroupingConstraints(m, curr_v2s=curr_v2s, v2var=v2var, enable_grouping_hints=True)

    # add optimization goal
    intra_edges, interface_edges = self.getIntraAndInterEdges(curr_v2s.keys())
    edge_costs = [m.add_var(var_type=INTEGER, name=f'intra_{e.name}') for e in intra_edges] \
        + [m.add_var(var_type=INTEGER, name=f'inter_{e.name}') for e in interface_edges]
    all_edges = intra_edges + interface_edges

    # note pos is different from slot_idx, becasue the x dimension is different from the y dimention
    # we will use |(y1 * 2 + y1) - (y2 * 2 + y2)| + |x1 - x2| to express the hamming distance
    pos_y = lambda v : v2var_y1[v] * 2 + v2var_y2[v]
    pos_x = lambda v : v2var_x[v]
    cost_y = lambda e : pos_y(e.src) - pos_y(e.dst)
    cost_x = lambda e : pos_x(e.src) - pos_x(e.dst)

    for e_cost_var, e in zip(edge_costs, all_edges):
      m += e_cost_var >= cost_y(e) + cost_x(e)
      m += e_cost_var >= -cost_y(e) + cost_x(e)
      m += e_cost_var >= cost_y(e) - cost_x(e)
      m += e_cost_var >= -cost_y(e) - cost_x(e)

    m.objective = minimize(xsum(edge_costs[i] * edge.width for i, edge in enumerate(all_edges) ) )

    _logger.info('Start ILP solver')
    # m.write('Coarse-Grained-Floorplan.lp')
    status = m.optimize(max_seconds=self.max_search_time)
    assert status == OptimizationStatus.OPTIMAL or status == OptimizationStatus.FEASIBLE, '8-way partioning failed!'

    # extract results
    next_s2v = {}
    next_v2s = {}

    for v in curr_v2s.keys():
      idx = int(slot_idx(v2var_y1[v].x,  v2var_y2[v].x, v2var_x[v].x))
      if slot_group[idx] not in next_s2v:
        next_s2v[slot_group[idx]] = []
      next_s2v[slot_group[idx]].append(v)
      next_v2s[v] = slot_group[idx]

    self.printFloorplan()

    # remove empty slots
    for slot in self.slot_manager.getActiveSlotsIncludeRouting():
      if slot not in next_s2v:
        self.slot_manager.removeSlotNonBlocking(slot.getName())

    self.s2v, self.v2s = next_s2v, next_v2s
    self.__initSlotToEdges()

    return

  # partition each slot separately. Used when the slots are already small
  def __separateTwoWayPartition(self, curr_s2v : Dict, curr_v2s : Dict, dir : str, external_v2s : Dict = {},
                                enable_grouping_hints = False,
                                exit_on_failure = True):
    next_s2v = defaultdict(list)
    next_v2s = {}

    # TODO: do we need to parallelize the loop?
    for slot, vertices in curr_s2v.items():
      individual_s2v = {slot : vertices}
      individual_v2s = {v : slot for v in vertices}

      # need external constraints
      external_v2s = {}
      for v in individual_v2s.keys():
        neighbors = v.getNeighborVertices()
        for n in neighbors:
          if n not in individual_v2s:
            external_v2s[n] = curr_v2s[n]

      # disable user constraints checkpoint in separate partition mode
      # because we only operate on a subset of the design, which may not cover
      # some of the user-specified modules. Just apply the constraints on available modules
      indi_next_s2v, indi_next_v2s = self.__twoWayPartitionWrapper(individual_s2v, individual_v2s, dir, external_v2s,
                                                                  enable_grouping_hints = enable_grouping_hints,
                                                                  check_user_constraints = False,
                                                                  exit_on_failure = exit_on_failure)

      if indi_next_s2v:
        next_s2v.update(indi_next_s2v)
      if indi_next_v2s:
        next_v2s.update(indi_next_v2s)

    return next_s2v, next_v2s

  # automatically adjust the max usage ratio to get a valid solution
  def __twoWayPartitionWrapper(self, curr_s2v : Dict, curr_v2s : Dict, dir : str, external_v2s : Dict = {},
                                enable_grouping_hints = True,
                                check_user_constraints=True,
                                exit_on_failure=True):
    init_delta = 0
    while 1:
      next_s2v, next_v2s = self.__twoWayPartition(curr_s2v, curr_v2s, dir, external_v2s, init_delta, enable_grouping_hints, check_user_constraints)
      if not next_s2v and not next_v2s:
        init_delta += self.step
        _logger.warning(f'use delta of {init_delta} to find valid solution')
        if self.max_usage_ratio + init_delta > self.MAX_USAGE_ALLOWED:
          _logger.critical('not likely there is a reasonable solution')
          if exit_on_failure:
            exit(1)
          else:
            _logger.critical('partition failed, returning unpartitioned mapping')
            return curr_s2v, curr_v2s
      else:
        return next_s2v, next_v2s

  # use iterative 2-way partitioning when there are lots of small functions
  # check_user_constraints: whether to check the modules in user-given constraints exist
  # set check_user_constraints to false in seperate partition of small slots
  def __twoWayPartition(self, curr_s2v : Dict, curr_v2s : Dict, dir : str, external_v2s : Dict, delta, enable_grouping_hints, check_user_constraints):
    assert set(map(type, curr_s2v.keys())) == {Slot}
    assert set(map(type, curr_v2s.keys())) == {Vertex}
    _logger.info(f'Start 2-way partitioning routine, pattern-based optimzation is {enable_grouping_hints}')

    m = Model()
    if not _logger.isEnabledFor(logging.DEBUG):
      m.verbose = 0

    v2var = self.__createILPVariables(m, curr_v2s=curr_v2s)

    self.__addOptGoal(m, curr_v2s=curr_v2s, external_v2s=external_v2s, v2var=v2var, dir=dir)

    # area constraints for each child slot
    self.__addAreaConstraints(m, curr_s2v=curr_s2v, v2var=v2var, dir=dir, delta=delta)

    self.__addUserConstraints(m, curr_v2s=curr_v2s, v2var=v2var, dir=dir, check_user_constraints=check_user_constraints)

    self.__addGroupingConstraints(m, curr_v2s=curr_v2s, v2var=v2var, enable_grouping_hints=enable_grouping_hints)

    _logger.info('Start ILP solver')
    # m.write('Coarse-Grained-Floorplan.lp')
    status = m.optimize(max_seconds=self.max_search_time)
    if status != OptimizationStatus.OPTIMAL and status != OptimizationStatus.FEASIBLE:
      _logger.warning('2-way partioning failed!')
      return {}, {}

    next_s2v, next_v2s = self.__getPartitionResult(m.num_solutions, curr_s2v=curr_s2v, v2var=v2var, dir=dir)

    return next_s2v, next_v2s

  def printFloorplan(self):
    _logger.info('Show current floorplan result:')

    for s, v_group in self.s2v.items():
      _logger.info(f'{s.getName()}:')
      for r in RESOURCE_TYPES:
        used = sum([v.area[r] for v in v_group])
        avail = s.getArea()[r]
        if avail:
          _logger.info(f'[{r}]: {used} / {avail} = {used/avail}')
        else:
          _logger.info(f'[{r}]: {used} used, 0 exists')
      for v in v_group:
        _logger.info(f'  Kernel: {v.name}')
      for e in self.s2e[s]:
        _logger.info(f'  FIFO: {e.name}')

    # wire information
    wire_length_list = []
    for e_list in self.s2e.values():
      for e in e_list:
        src_slot = self.v2s[e.src]
        dst_slot = self.v2s[e.dst]
        length = abs(src_slot.getPositionY() - dst_slot.getPositionY()) + \
                 abs(src_slot.getPositionX() - dst_slot.getPositionX())
        wire_length_list.append(length)

    _logger.info(f'total wire length: {sum(wire_length_list)}')
    if len(wire_length_list) > 1:
      _logger.info(f'variance of wire length: {statistics.variance(wire_length_list)}')

    # SLR crossing information
    slr_crossing = [0] * (self.board.SLR_NUM-1)
    for e_list in self.s2e.values():
      for e in e_list:
        src_slr = self.v2s[e.src].getSLR()
        dst_slr = self.v2s[e.dst].getSLR()
        idx_small = min(src_slr, dst_slr)
        idx_large = max(src_slr, dst_slr)
        for i in range(idx_small, idx_large):
          slr_crossing[i] += e.width

    for i in range(self.board.SLR_NUM-1):
      _logger.info(f'SLR boundary {i} - {i+1} has {slr_crossing[i]} crossings')

  def getUtilization(self) -> Dict[Slot, Dict[str, float]]:
    util = {}
    for s, v_group in self.s2v.items():
      assert v_group, f'empty slot should not exist: {s.getName()}'
      util[s] = {}
      for r in RESOURCE_TYPES:
        used = sum([v.area[r] for v in v_group])
        avail = s.getArea()[r]
        # a slot may not contain some type of resources
        if avail:
          util[s][r] = used/avail

    return util

  # obtain the edges that are inside the given slots and the edges between the given slots and the other slots
  def getIntraAndInterEdges(self, v_group):
    second_visited_edges = set()
    first_visited_edges = set()

    # if an edge is visited twice, then it is entirely within the target slots
    # if an edge is visited only once, then it is between the target slots and the remaining slots
    for v in v_group:
      for e in v.getEdges():
        if e in first_visited_edges:
          second_visited_edges.add(e)
          first_visited_edges.remove(e)
        else:
          first_visited_edges.add(e)

          # double check that an edge will not be visited a 3rd time
          assert e not in second_visited_edges

    interface_edges = list(first_visited_edges)
    intra_edges = list(second_visited_edges)

    return intra_edges, interface_edges

  def coarseGrainedFloorplan(self):
    init_s2v, init_v2s = self.__getInitialSlotToVerticesMapping()
    iter1_s2v, iter1_v2s = self.__twoWayPartitionWrapper(init_s2v, init_v2s, Dir.horizontal) # based on die boundary

    iter2_s2v, iter2_v2s = self.__twoWayPartitionWrapper(iter1_s2v, iter1_v2s, Dir.horizontal) # based on die boundary

    self.s2v, self.v2s = self.__twoWayPartitionWrapper(iter2_s2v, iter2_v2s, Dir.vertical) # based on ddr ctrl in the middle

    self.__initSlotToEdges()
    self.printFloorplan()

  # do not use pattern-based clustering
  def naiveFineGrainedFloorplan(self):
    init_s2v, init_v2s = self.__getInitialSlotToVerticesMapping()
    iter1_s2v, iter1_v2s = self.__twoWayPartitionWrapper(init_s2v, init_v2s, Dir.horizontal, enable_grouping_hints=False) # based on die boundary

    iter2_s2v, iter2_v2s = self.__twoWayPartitionWrapper(iter1_s2v, iter1_v2s, Dir.horizontal, enable_grouping_hints=False) # based on die boundary

    iter3_s2v, iter3_v2s = self.__twoWayPartitionWrapper(iter2_s2v, iter2_v2s, Dir.vertical, enable_grouping_hints=False) # based on die boundary

    iter4_s2v, iter4_v2s = self.__separateTwoWayPartition(iter3_s2v, iter3_v2s, Dir.horizontal, enable_grouping_hints=False) # based on die boundary

    self.s2v, self.v2s = self.__separateTwoWayPartition(iter4_s2v, iter4_v2s, Dir.vertical, enable_grouping_hints=False) # based on ddr ctrl in the middle

    self.__initSlotToEdges()
    self.printFloorplan()

  def floorplanVHHvh(self):
    """
    switch the order of partitioning. Seems that this one is much faster than other order.
    """
    init_s2v, init_v2s = self.__getInitialSlotToVerticesMapping()
    iter1_s2v, iter1_v2s = self.__twoWayPartitionWrapper(init_s2v, init_v2s, Dir.vertical) # based on die boundary

    iter2_s2v, iter2_v2s = self.__twoWayPartitionWrapper(iter1_s2v, iter1_v2s, Dir.horizontal) # based on die boundary

    iter3_s2v, iter3_v2s = self.__twoWayPartitionWrapper(iter2_s2v, iter2_v2s, Dir.horizontal) # based on ddr ctrl in the middle

    iter4_s2v, iter4_v2s = self.__separateTwoWayPartition(iter3_s2v, iter3_v2s, Dir.vertical, enable_grouping_hints=False) # 4 CR

    iter5_s2v, iter5_v2s = self.__separateTwoWayPartition(iter4_s2v, iter4_v2s, Dir.horizontal, enable_grouping_hints=False) # 8 CR

    v_list = self.graph.getAllVertices()
    v_name_to_v = {v.name : v for v in v_list}
    grouping_list: List[List[Vertex]] = [
      [v_name_to_v[v_name] for v_name in grouping] \
        for grouping in self.grouping_constraints
    ]

    # FIXME: a temp hack to allow the legalizer to use all slots
    all_slot_list = self.slot_manager.getAllTwoByTwoCRSlots()
    self.s2v, self.v2s = AutoLegalizer(iter5_v2s, grouping_list, all_slot_list)

    # FIXME: a temp hack to register all used slots as compute slots
    for slot in self.s2v.keys():
      self.slot_manager.createSlot(slot.getName())

    self.__initSlotToEdges()
    self.printFloorplan()

  def patternBasedFineGrainedFloorplan(self):
    init_s2v, init_v2s = self.__getInitialSlotToVerticesMapping()
    iter1_s2v, iter1_v2s = self.__twoWayPartitionWrapper(init_s2v, init_v2s, Dir.horizontal) # based on die boundary

    iter2_s2v, iter2_v2s = self.__twoWayPartitionWrapper(iter1_s2v, iter1_v2s, Dir.horizontal) # based on die boundary

    iter3_s2v, iter3_v2s = self.__twoWayPartitionWrapper(iter2_s2v, iter2_v2s, Dir.vertical) # based on ddr ctrl in the middle

    iter4_s2v, iter4_v2s = self.__separateTwoWayPartition(iter3_s2v, iter3_v2s, Dir.horizontal, enable_grouping_hints=False) # 8 CR

    self.s2v, self.v2s = self.__separateTwoWayPartition(iter4_s2v, iter4_v2s, Dir.vertical, enable_grouping_hints=False) # 4 CR

    self.__initSlotToEdges()
    self.printFloorplan()

  def hetero4CRFloorplan(self):
    init_s2v, init_v2s = self.__getInitialSlotToVerticesMapping()
    iter1_s2v, iter1_v2s = self.__twoWayPartitionWrapper(init_s2v, init_v2s, Dir.horizontal) # based on die boundary

    iter2_s2v, iter2_v2s = self.__twoWayPartitionWrapper(iter1_s2v, iter1_v2s, Dir.horizontal) # based on die boundary

    iter3_s2v, iter3_v2s = self.__twoWayPartitionWrapper(iter2_s2v, iter2_v2s, Dir.vertical) # based on ddr ctrl in the middle

    iter4_s2v, iter4_v2s = self.__separateTwoWayPartition(iter3_s2v, iter3_v2s, Dir.vertical, enable_grouping_hints=False) # 8 CR

    # in the last partition, go as small as possible but do not enforce
    # allow heterogeneous slot sizes
    self.s2v, self.v2s = self.__separateTwoWayPartition(iter4_s2v, iter4_v2s, Dir.horizontal, enable_grouping_hints=False, exit_on_failure=False) # 4 CR

    self.__initSlotToEdges()
    self.printFloorplan()

  def naiveTwoCRGranularityFloorplan(self):
    init_s2v, init_v2s = self.__getInitialSlotToVerticesMapping()
    iter1_s2v, iter1_v2s = self.__twoWayPartitionWrapper(init_s2v, init_v2s, Dir.horizontal) # based on die boundary

    iter2_s2v, iter2_v2s = self.__twoWayPartitionWrapper(iter1_s2v, iter1_v2s, Dir.horizontal) # based on die boundary

    iter3_s2v, iter3_v2s = self.__twoWayPartitionWrapper(iter2_s2v, iter2_v2s, Dir.horizontal) # based on die boundary

    iter4_s2v, iter4_v2s = self.__twoWayPartitionWrapper(iter3_s2v, iter3_v2s, Dir.vertical) # based on die boundary

    iter5_s2v, iter5_v2s = self.__twoWayPartitionWrapper(iter4_s2v, iter4_v2s, Dir.vertical) # based on die boundary

    self.s2v, self.v2s = self.__twoWayPartitionWrapper(iter5_s2v, iter5_v2s, Dir.vertical) # based on ddr ctrl in the middle

    self.__initSlotToEdges()
    self.printFloorplan()

  def getSlotToVertices(self):
    return self.s2v

  def getSlotToEdges(self):
    return self.s2e

  def getVertexToSlot(self):
    return self.v2s

  def getSlotNameToVertexNames(self):
    s_name_2_v_names = {}
    for slot, v_group in self.s2v.items():
      s_name_2_v_names[slot.getName()] = {v.type : v.name for v in v_group}
    return s_name_2_v_names

  def getVertexNameToSlot(self):
    return {v.name: s for v, s in self.v2s.items()}

  def getEdgeNameToSlot(self):
    e_name_to_s = {}
    for slot, e_list in self.s2e.items():
      for e in e_list:
        e_name_to_s[e.name] = slot

    return e_name_to_s

  def getSlotNameToEdgeNames(self):
    s_name_2_e_names = {}
    for slot, e_group in self.s2e.items():
      s_name_2_e_names[slot.getName()] = [e.name for e in e_group]
    return s_name_2_e_names


import logging

from typing import Dict, List, Optional
from autobridge.Opt.DataflowGraph import Vertex, Edge, DataflowGraph
from autobridge.Opt.Slot import Slot
from autobridge.Opt.SlotManager import SlotManager, Dir
from autobridge.Floorplan.Bipartition import Bipartition
from autobridge.Floorplan.LegalizeFloorplan import legalize_floorplan
from autobridge.Floorplan.Utilities import log_resource_utilization
_logger = logging.getLogger().getChild(__name__)


def iterative_bipartition(
  init_v2s: Dict[Vertex, Slot],
  slot_manager: SlotManager,
  grouping_constraints: List[List[Vertex]],
  pre_assignments: Dict[Vertex, Slot],
  partition_order: List[Dir],
  ref_usage_ratio: float,
) -> Optional[Dict[Vertex, Slot]]:
  """
  first iteratively partition the graph. Allow the usage ratio to increase in the process
  Globally adjust the floorplan in the end to meet the usage ratio
  """
  floorplaner = Bipartition(
    init_v2s,
    grouping_constraints,
    pre_assignments,
    slot_manager,
  )

  # performs multiple iteration of partitioning
  curr_v2s = {}
  for idx, split_dir in enumerate(partition_order):
    _logger.info(f'partition #{idx+1} in the direction: {split_dir}')
    post_partition_v2s = floorplaner.get_bipartition_adjust_ratio(split_dir, ref_usage_ratio)

    # globally adjust the floorplan
    legalized_v2s = legalize_floorplan(
      post_partition_v2s, 
      slot_manager, 
      grouping_constraints, 
      pre_assignments, 
      partition_order[:idx+1],
      max_usage_ratio=0.85,
    )

    if not legalized_v2s:
      _logger.info(f'The current bi-partition iteration failed. Skip the remaining partition iterations.')
      _logger.info(f'WARNING: some functions in the design are too large. Try to break them into smaller functions.')
      return curr_v2s
    else:
      curr_v2s = legalized_v2s

    floorplaner.set_curr_v2s(curr_v2s)

    log_resource_utilization(curr_v2s)

  return curr_v2s



import collections


class ProgramJsonManager:

  def __init__(self, fifo_section, v_name_to_v_module, v_module_to_area):
    self.fifo_section = fifo_section
    self.v_name_to_v_module = v_name_to_v_module
    self.v_module_to_area = v_module_to_area
    self.vertex_total_area = collections.defaultdict(int)
    for v_name, v_module in self.v_name_to_v_module.items():
      for item, count in self.v_module_to_area[v_module].items():
        self.vertex_total_area[item] += count

  def getFIFOSection(self):
    return self.fifo_section

  def getVNameToModule(self):
    return self.v_name_to_v_module

  def getAreaOfModule(self, v_module):
    return self.v_module_to_area[v_module]

  def getVertexTotalArea(self):
    return self.vertex_total_area



#! /usr/bin/python3.6
import logging
from collections import defaultdict
from typing import Dict, Iterable, Iterator, List, Union
import re
import json

class Slot:
  def __init__(self, board, pblock : str):
    self.board = board

    assert 'COARSE_' not in pblock
    
    match = re.search(r'^CLOCKREGION_X(\d+)Y(\d+)[ ]*:[ ]*CLOCKREGION_X(\d+)Y(\d+)$', pblock)
    if not match:
      match = re.search(r'^CR_X(\d+)Y(\d+)[ ]*_To_[ ]*CR_X(\d+)Y(\d+)$', pblock)
    assert match, f'incorrect pblock {pblock}'

    # DO NOT convert CR coordinate to boundary intersect coordinates
    # stick to vivado convention
    self.down_left_x = int(match.group(1))
    self.down_left_y = int(match.group(2))
    self.up_right_x =  int(match.group(3))
    self.up_right_y =  int(match.group(4))

    assert self.down_left_x < 100
    assert self.down_left_y < 100
    assert self.up_right_x < 100
    assert self.up_right_y < 100

    self.area = self.board.getArea(f'CLOCKREGION_X{self.down_left_x}Y{self.down_left_y}:CLOCKREGION_X{self.up_right_x}Y{self.up_right_y}')

  def getName(self):
    # need to convert back to CR coordinates
    return f'CLOCKREGION_X{self.down_left_x}Y{self.down_left_y}:CLOCKREGION_X{self.up_right_x}Y{self.up_right_y}'

  def getOrigUpRightX(self):
    return self.up_right_x
  def getOrigUpRightY(self):
    return self.up_right_y
  def getOrigDownLeftX(self):
    return self.down_left_x
  def getOrigDownLeftY(self):
    return self.down_left_y

  def getRTLModuleName(self):
    return f'CR_X{self.down_left_x}Y{self.down_left_y}_To_CR_X{self.up_right_x}Y{self.up_right_y}'

  def __key(self):
    return (str(self.down_left_x).zfill(3),
          str(self.down_left_y).zfill(3),
          str(self.up_right_x).zfill(3),
          str(self.up_right_y).zfill(3))

  def __hash__(self):
    return hash(self.__key())

  def __eq__(self, other):
    if isinstance(other, Slot):
      return self.__key() == other.__key()
    assert False, 'comparing Slot to a different class'

  def getArea(self):
    return self.area
  
  # use the middle point as the position of the slot. Check the results have no fractional part
  # biased towards the larger side
  def getPositionX(self):
    return int((self.down_left_x + self.up_right_x + 1) / 2)

  # biased towards the ceiling side
  def getPositionY(self):
    return int((self.down_left_y + self.up_right_y + 1) / 2)

  def getLenX(self):
    return self.up_right_x - self.down_left_x + 1

  def getLenY(self):
    return self.up_right_y - self.down_left_y + 1
  
  # 1/4 from the lower end
  def getQuarterPositionX(self):
    return int(self.down_left_x + (self.up_right_x - self.down_left_x + 1) / 4)

  def getQuarterPositionY(self):
    return int(self.down_left_y + (self.up_right_y - self.down_left_y + 1) / 4)

  def getHalfLenX(self):
    return int((self.up_right_x - self.down_left_x + 1) / 2)

  def getHalfLenY(self):
    return int((self.up_right_y - self.down_left_y + 1) / 2)

  #                  |-------| u_r_x, u_r_y
  #                  |       |
  #                  |  up   |
  #                  |       |
  #                  |-------| u_r_x, mid_y   
  #                  |       |
  #                  |  bot  |
  #                  |       |
  #   d_l_x, d_l_y   |-------|
  def getBottomChildSlotName(self):
    assert self.down_left_x != self.up_right_x or \
      self.down_left_y != self.up_right_y, 'Cannot split a single CR'

    mid_y = self.getPositionY()
    down_left_cr = f'CLOCKREGION_X{self.down_left_x }Y{self.down_left_y}'
    up_right_cr  = f'CLOCKREGION_X{self.up_right_x}Y{mid_y-1}'
    return f'{down_left_cr}:{up_right_cr}'
  
  def getUpChildSlotName(self):
    assert self.down_left_x != self.up_right_x or \
      self.down_left_y != self.up_right_y, 'Cannot split a single CR'

    mid_y = self.getPositionY()
    down_left_cr = f'CLOCKREGION_X{self.down_left_x }Y{mid_y}'
    up_right_cr  = f'CLOCKREGION_X{self.up_right_x}Y{self.up_right_y}'
    return f'{down_left_cr}:{up_right_cr}'

  #                  mid_x, u_r_y
  #               |---------|---------| u_r_x, u_r_y
  #               |         |         |
  #               |  L      |      R  |
  #               |         |         |
  #  d_l_x, d_l_y |---------|---------|
  #                    mid_x, d_l_y
  #     
  def getLeftChildSlotName(self):
    assert self.down_left_x != self.up_right_x or \
      self.down_left_y != self.up_right_y, 'Cannot split a single CR'

    mid_x = self.getPositionX()
    down_left_cr = f'CLOCKREGION_X{self.down_left_x}Y{self.down_left_y}'
    up_right_cr  = f'CLOCKREGION_X{mid_x - 1       }Y{self.up_right_y}'
    return f'{down_left_cr}:{up_right_cr}'

  def getRightChildSlotName(self):
    assert self.down_left_x != self.up_right_x or \
      self.down_left_y != self.up_right_y, 'Cannot split a single CR'

    mid_x = self.getPositionX()
    down_left_cr = f'CLOCKREGION_X{mid_x            }Y{self.down_left_y}'
    up_right_cr  = f'CLOCKREGION_X{self.up_right_x}Y{self.up_right_y}'
    return f'{down_left_cr}:{up_right_cr}'

  def containsChildSlot(self, target) -> bool:
    return target.down_left_x >= self.down_left_x \
      and  target.down_left_y >= self.down_left_y \
      and  target.up_right_x  <= self.up_right_x  \
      and  target.up_right_y  <= self.up_right_y  

  def getNeighborSlotName(self, dir):
    """get the neighbor slot at the given direction of the equal size"""
    len_x = self.up_right_x - self.down_left_x + 1
    len_y = self.up_right_y - self.down_left_y + 1

    if dir == 'UP':
      return f'CLOCKREGION_X{self.down_left_x}Y{self.down_left_y+len_y}:CLOCKREGION_X{self.up_right_x}Y{self.up_right_y+len_y}'
    elif dir == 'DOWN':
      return f'CLOCKREGION_X{self.down_left_x}Y{self.down_left_y-len_y}:CLOCKREGION_X{self.up_right_x}Y{self.up_right_y-len_y}'
    elif dir == 'RIGHT':
      return f'CLOCKREGION_X{self.down_left_x+len_x}Y{self.down_left_y}:CLOCKREGION_X{self.up_right_x+len_x}Y{self.up_right_y}'
    elif dir == 'LEFT':
      return f'CLOCKREGION_X{self.down_left_x-len_x}Y{self.down_left_y}:CLOCKREGION_X{self.up_right_x-len_x}Y{self.up_right_y}'
    else: assert False

  def getSLR(self):
    """ return which SRL the slot is in, starting from 0 """
    num_rows_per_slr = self.board.CR_NUM_VERTICAL_PER_SLR
    for i in range(self.board.SLR_NUM):
      if self.down_left_y >= i * num_rows_per_slr \
         and self.up_right_y <= (i+1) * num_rows_per_slr - 1:
        return i
    
    assert False, f'the current slot {self.getName()} is beyond 1 SLR'
    return None

  def getBoundarySegmentName(self, dir):
    if dir == 'UP':
      return f'Boundary_X{self.down_left_x}Y{self.up_right_y+1}_To_X{self.up_right_x+1}Y{self.up_right_y+1}'
    elif dir == 'DOWN':
      return f'Boundary_X{self.down_left_x}Y{self.down_left_y}_To_X{self.up_right_x+1}Y{self.down_left_y}'
    elif dir == 'RIGHT':
      return f'Boundary_X{self.up_right_x+1}Y{self.down_left_y}_To_X{self.up_right_x+1}Y{self.up_right_y+1}'
    elif dir == 'LEFT':
      return f'Boundary_X{self.down_left_x}Y{self.down_left_y}_To_X{self.down_left_x}Y{self.up_right_y+1}'
    else:
      assert False

  def getDistance(self, other: 'Slot') -> int:
    """
    get the manhatten dist from another slot
    """
    dist_x = abs(self.getPositionX() - other.getPositionX()) 
    dist_y = abs(self.getPositionY() - other.getPositionY())
    return dist_x + dist_y

  #------------- For TAPA ------------- #
  def isToTheLeftOf(self, other: 'Slot') -> bool:
    return (self.down_left_y == other.down_left_y and
            self.up_right_y == other.up_right_y and
            self.up_right_x+1 == other.down_left_x)

  def isToTheRightOf(self, other: 'Slot') -> bool:
    return other.isToTheLeftOf(self)

  def isAbove(self, other: 'Slot') -> bool:
    return (self.down_left_x == other.down_left_x and
            self.up_right_x == other.up_right_x and
            self.down_left_y == other.up_right_y+1)

  def isBelow(self, other: 'Slot') -> bool:
    return other.isAbove(self)

  @property
  def pblock_name(self) -> str:
    return self.getRTLModuleName()

  @property
  def pblock_tcl(self) -> str:
    """ 
    remove the overlaps with vitis IPs 
    must start with a leading '\n', or the tapa side will error
    """
    return '\n' + f'# begin defining a slot\n' + '\n'.join(self.board.getSlotPblockTcl(self)) + '\n'
  #------------------------------------ #

# For TAPA. To be replaced by GlobalRouting
class Topology:

  def __init__(self, slots: Iterable[Slot]):
    self.slots = tuple(slots)
    self.adjancency: Dict[str, Dict[str, str]] = defaultdict(dict)
    for slot1 in self.slots:
      for slot2 in self.slots:
        if slot2.isToTheLeftOf(slot1):
          self.adjancency[slot1.pblock_name]['left'] = slot2.pblock_name
        elif slot2.isToTheRightOf(slot1):
          self.adjancency[slot1.pblock_name]['right'] = slot2.pblock_name
        elif slot2.isAbove(slot1):
          self.adjancency[slot1.pblock_name]['up'] = slot2.pblock_name
        elif slot2.isBelow(slot1):
          self.adjancency[slot1.pblock_name]['down'] = slot2.pblock_name

  def getNeighborOf(self, name: str, direction: str) -> str:
    """Get neighbor of `name` on `direction`. Empty if none exists."""
    return self.adjancency.get(name, {}).get(direction, '')

  def yieldPaths(
      self,
      direction: str,
      init: List[str],
  ) -> Iterator[List[str]]:
    """Yield connections on `direction` starting from `init`.

    Arguments:
      directions: One of 'up', 'down', 'left', or 'right'.
      init: Initial connection.

    Yields:
      Connections with prefix `init` on `direction` (excluding `init` itself).
    """
    name = init[-1]
    neighbors = init[:]
    while True:
      name = self.getNeighborOf(name, direction)
      if not name:
        break
      neighbors.append(name)
      yield neighbors[:]

  def getTopologyOf(self, slot: Slot) -> Dict[str, Union[str, List[str]]]:
    # list of connections; each connection is a list starting from `slot` to the
    # slots that are connected to `slot`
    paths: List[List[str]] = []
    for direction in 'left', 'right':
      for path in self.yieldPaths(direction, [slot.pblock_name]):
        paths.append(path)
        for next_direction in 'up', 'down':
          paths.extend(self.yieldPaths(next_direction, path))
    for direction in 'up', 'down':
      for path in self.yieldPaths(direction, [slot.pblock_name]):
        paths.append(path)
        for next_direction in 'left', 'right':
          paths.extend(self.yieldPaths(next_direction, path))

    # p[-1] is the destination, p[0] is the source, neither of them is required
    # in the value
    topology: Dict[str, Union[str, List[str]]] = {p[-1]: p[1:-1] for p in paths}
    topology['tcl'] = slot.pblock_tcl
    return topology



#! /usr/bin/python3.6
from collections import defaultdict
from pyverilog.vparser.parser import parse as rtl_parse
import pyverilog.vparser.ast as ast
from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
import re
import logging
import json

class AXIConnectionParser:
  def __init__(self, top_rtl_path):
    self.top_rtl_path = top_rtl_path
    self.top_module_ast, directives = rtl_parse([top_rtl_path]) 
    self.top_m_axi_port_to_io_module = self.__getPortToModule()
    self.s_axi_module = self.__initSAXIModule()
    # print(json.dumps(self.top_m_axi_port_to_io_module, indent=2))

  def __getTopMAXIs(self):
    top = open(self.top_rtl_path, 'r').read()
    top_axi_list = re.findall(r'input[ ]+m_axi_(.*)_ARREADY', top)
    return top_axi_list

  def __DFS(self, node, filter_func):
    if filter_func(node):
      try:
        logging.debug(f'visit node {node.name}')
      except:
        pass
        # logging.debug(f'node in line {node.lineno} has no name')
      yield node
    for c in node.children():
      yield from self.__DFS(c, filter_func)

  def __isVertexNode(self, node):
    return isinstance(node, ast.Instance) and 'fifo' not in node.module

  def traverseVertexInAST(self):
    yield from self.__DFS(self.top_module_ast, self.__isVertexNode)

  # extract the modules directly connected to the top M_AXI interfaces
  def __getPortToModule(self):
    m_axi_list = self.__getTopMAXIs()

    top_m_axi_port_to_io_module = {}

    for v_node in self.traverseVertexInAST():
      interface_wire_set = set()
      for portarg in v_node.portlist:
        if not isinstance(portarg.argname, ast.Identifier): continue
        interface_wire_set.add(portarg.argname.name)
      # interface_wire_set = set([portarg.argname.name  for portarg in v_node.portlist if not isinstance(portarg.argname, ast.Identifier)] )
      for axi_name in m_axi_list:
        if f'm_axi_{axi_name}_ARREADY' in interface_wire_set:
          top_m_axi_port_to_io_module[axi_name] = (v_node.module, v_node.name)
    
    assert len(top_m_axi_port_to_io_module) == len(m_axi_list)
    return top_m_axi_port_to_io_module

  def getIOModuleNameOfAXI(self, axi_name):
    return self.top_m_axi_port_to_io_module[axi_name][1] # get name 

  def getAllAXIModules(self) -> list:
    axi_modules = list(self.top_m_axi_port_to_io_module.values())
    axi_modules.append(self.s_axi_module)

    return axi_modules
  
  def __initSAXIModule(self) -> list:
    for v_node in self.traverseVertexInAST():
      for portarg in v_node.portlist:
        if not isinstance(portarg.argname, ast.Identifier): continue
        if 's_axi_control' in portarg.argname.name:
          return (v_node.module, v_node.name)

  def getSAXIName(self):
    return self.s_axi_module[1]

if __name__ == '__main__':
  top_rtl_path = '/home/einsx7/auto-parallel/src/benchmark/SSSP-tapa/hdl/SSSP_SSSP.v'
  AXIConnectionParser(top_rtl_path)



#! /usr/bin/python3.6

import logging
import re
from os.path import isfile, isdir
from collections import defaultdict

class HLSProjectManager:
  def __init__(
      self, 
      top_func_name,
      hls_prj_path, 
      hls_solution_name='solution'):
    self.top_func_name = top_func_name
    self.hls_prj_path = hls_prj_path
    self.hls_solution_name = hls_solution_name
    self.area_map = {} # name -> area
    self.__checker()

    self.getAreaOfAllInst()

  def __checker(self):
    # rtl name should contain not the file extension
    assert self.top_func_name[-2:] != '.v'

  def getCsynthReportDir(self):
    ans = f'{self.hls_prj_path}/{self.hls_solution_name}/syn/report/'
    assert isdir(ans), ans
    return ans

  def getScheReportDir(self):
    ans =  f'{self.hls_prj_path}/{self.hls_solution_name}/.autopilot/db/'
    assert isdir(ans), ans
    return ans

  def getRTLDir(self):
    ans = f'{self.hls_prj_path}/{self.hls_solution_name}/syn/verilog/'
    assert isdir(ans), ans
    return ans

  def getTopRTLPath(self):
    opt1 = f'{self.hls_prj_path}/{self.hls_solution_name}/syn/verilog/{self.top_func_name}.v'
    opt2 = f'{self.hls_prj_path}/{self.hls_solution_name}/syn/verilog/{self.top_func_name}_{self.top_func_name}.v'
    if isfile(opt1):
      return opt1
    elif isfile(opt2):
      return opt2
    else:
      assert False, f'cannot find the RTL file for {self.top_func_name}'  

  # kernel0 or kernel0_kernel0
  def getTopModuleName(self):
    opt1 = rf'{self.top_func_name}'
    opt2 = rf'{self.top_func_name}_{self.top_func_name}'
    path = self.getTopRTLPath()
    if re.search(opt2, path):
      return opt2
    elif re.search(opt1, path):
      return opt1
    else:
      assert False, f'cannot find the top RTL module name'   

  def getScheReportFromModuleType(self, mod_type):
    opt1 = self.getScheReportDir() + f'/{mod_type}' + '.verbose.sched.rpt'
    opt2 = self.getScheReportDir() + f'/{mod_type[len(self.top_func_name)+1:]}' + '.verbose.sched.rpt'
    
    if isfile(opt1):
      return opt1
    elif isfile(opt2):
      return opt2
    else:
      assert False, f'cannot find the schedule report for {mod_type}'

  def getHLSReportFromModuleType(self, mod_type):
    opt1 = self.getCsynthReportDir() + f'/{mod_type}' + '_csynth.rpt'
    opt2 = self.getCsynthReportDir() + f'/{mod_type[len(self.top_func_name)+1:]}' + '_csynth.rpt'
    
    if isfile(opt1):
      return opt1
    elif isfile(opt2):
      return opt2
    else:
      assert False, f'cannot find the HLS report for {mod_type}'    

  def getAreaOfAllInst(self):
    top_rpt_addr = self.getHLSReportFromModuleType(self.top_func_name)
    rpt = open(top_rpt_addr, 'r')
    line = ''

    while r'Utilization Estimates' not in line:
      line = rpt.readline()

    # ......
    # + Detail: 
    # * Instance: 
    # +-------------------------------------+------------------------------------+---------+-------+------+------+-----+
    # |               Instance              |               Module               | BRAM_18K| DSP48E|  FF  |  LUT | URAM|
    # +-------------------------------------+------------------------------------+---------+-------+------+------+-----+
    # |A_IO_L1_in_0_0_U0                    |A_IO_L1_in_0_0_s                    |        0|      0|   429|   334|    0|
    # |A_IO_L1_in_0_10_U0                   |A_IO_L1_in_0_10_s                   |        0|      0|   425|   334|    0|
    # ......

    while r'+ Detail:' not in line:
      line = rpt.readline()

    line = rpt.readline()   # * Instance: 
    assert r'* Instance:' in line, line
    line = rpt.readline()   # +-------------
    assert r'+-------' in line, line
    line = rpt.readline()   # |               Instance  
    assert re.search(r'[| ]*Instance[ ]*[| ]*Module[ ]*[| ]*BRAM_18K[ |]+DSP[48E |]+FF[ |]+LUT[ |]+URAM', line), line
    line = rpt.readline()   # +----------
    assert r'+-------' in line, line

    line = rpt.readline()   # +----------
    while r'+-------' not in line:
      #                   |   (inst )     |    (mod )     |    (bram)   |    (dsp)    |    (ff )    |   (lut)     |    (uram)
      match = re.search(r'[| ]*([^ ]+)[ ]*[| ]*([^ ]+)[ ]*[| ]*(\d+)[ ]*[| ]*(\d+)[ ]*[| ]*(\d+)[ ]*[| ]*(\d+)[ ]*[| ]*(\d+)', line)
      assert match, line

      inst_name = match.group(1)
      mod_type = match.group(2)
      bram = int(match.group(3))
      dsp  = int(match.group(4))
      ff   = int(match.group(5))
      lut  = int(match.group(6))
      uram = int(match.group(7))
      self.area_map[mod_type] = {'BRAM':bram, 'DSP':dsp, 'FF':ff, 'LUT':lut, 'URAM':uram}

      line = rpt.readline() 
    rpt.close()
    
  def getAreaFromModuleType(self, mod_type):
    # if '_axi' in mod_type:
    #   if mod_type not in self.area_map:
    #     assert re.search(f'{self.top_func_name}_{self.top_func_name}', mod_type), mod_type
    #     mod_type = mod_type[len(self.top_func_name)+1:]
    
    opt1 = mod_type
    opt2 = mod_type[len(self.top_func_name)+1:]
    if opt1 in self.area_map: 
      return self.area_map[opt1]
    elif opt2 in self.area_map:
      return self.area_map[opt2]
    else:
      return self.getAreaBasedOnIndividualReport(mod_type)

  def getAreaBasedOnIndividualReport(self, mod_type):
    logging.debug(f'Area information of instance {mod_type} not found in report for the top func, but has its own report.')

    rpt_addr = self.getHLSReportFromModuleType(mod_type)
    rpt = open(rpt_addr, 'r')

    for line in rpt:
      if ('Utilization Estimates' in line):
        for line in rpt:
          if ('Name' in line):
            assert re.search(r'BRAM[_]18K[ |]*DSP[48E |]*FF[ |]*LUT[ |]*URAM', line), f'HLS has changed the item order in reports! {rpt_addr} : {line}'

          if ('Total' in line):
            x = re.findall(r'\d+', line)
            return {'BRAM':int(x[0]), 'DSP':int(x[1]), 'FF':int(x[2]), 'LUT':int(x[3]), 'URAM':int(x[4])}

    assert False, 'Error in parsing the HLS report'

  def getTotalArea(self):
    return self.getAreaBasedOnIndividualReport(self.top_func_name)


#! /usr/bin/python3.6
from collections import defaultdict
from pyverilog.vparser.parser import parse as rtl_parse
import pyverilog.vparser.ast as ast
from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
import re
import logging
import math

class TopRTLParser:

  fifo_dout_format = '([^ ]*[^_])_+dout'
  fifo_din_format = '([^ ]*[^_])_+din'

  def __init__(self, top_rtl_path):
    self.top_rtl_path = top_rtl_path
    self.top_module_ast, directives = rtl_parse([top_rtl_path]) 
    self.__checker()

    self.mod_to_fifo_in = defaultdict(list)
    self.mod_to_fifo_out = defaultdict(list)

    self.wire_to_fifo_name = {} # str -> str
    self.fifo_name_to_wires = defaultdict(list) # fifo -> interface wires
    self.fifo_name_to_outbound_wires = defaultdict(list)
    self.fifo_name_to_inbound_wires = defaultdict(list)

    # CRITICAL:
    # if two vertices are connected by direct wire
    # one wire will map to 2 vertices
    self.wire_to_v_name = {} # str -> str
    
    self.v_name_to_wires = defaultdict(list) # vertex -> interface wires
    self.inst_name_to_rtl = {}
    self.reg_wire_name_to_width = {} # from name to full declaration (with width, etc) 
    self.io_name_to_width = {}
    self.io_name_to_dir = {}
    self.all_decl_except_io = []
    self.codegen = ASTCodeGenerator()
    self.ap_done_v_name_to_wire = {} # ap_done module name -> wire name
    self.ap_ready_v_name_to_wire = {}
    self.param_to_value_str = {}
    self.e_name_to_ast_node = {}

    # record which vertices have direct wire connection
    self.grouping_constraints = set()

    self.__initWireToFIFOMapping()
    self.__initWireToVertexMapping()
    self.__initFIFOListOfModuleInst()
    self.__initRTLOfAllInsts()
    self.__initParamToValueStr()
    self.__initDeclList()
    self.__initApDoneSources()
    self.__initApReadySources()
    
    # for mod, ap_done in self.ap_done_v_name_to_wire.items():
    #   print(f'{ap_done} -> {mod}')

  # 1. no start_for FIFOs
  # 2. each inst has different name
  def __checker(self):
    for node in self.__DFS(self.top_module_ast, lambda node : isinstance(node, ast.Instance)):
      assert 'start_for' not in node.name, f'Found start FIFOs: {self.top_rtl_path} : {node.name}' 

    names = set()
    for node in self.__DFS(self.top_module_ast, lambda node : isinstance(node, ast.Instance)):
      if node.name not in names:
        names.add(node.name)
      else:
        assert False, f'Found duplicated name for instance {node.name}'

  def __DFS(self, node, filter_func):
    if filter_func(node):
      try:
        logging.debug(f'visit node {node.name}')
      except:
        pass
        # logging.debug(f'node in line {node.lineno} has no name')
      yield node
    for c in node.children():
      yield from self.__DFS(c, filter_func)

  def __initRTLOfAllInsts(self):
    for v_inst_list in self.__DFS(self.top_module_ast, self.__isVertexInstanceList):
      assert len(v_inst_list.instances) == 1, f'unsupported RTL coding style at line {v_inst_list.lineno}'
      v_node = v_inst_list.instances[0]
      self.inst_name_to_rtl[v_node.name] = self.codegen.visit(v_inst_list)

    for e_inst_list in self.__DFS(self.top_module_ast, self.__isEdgeInstanceList):
      assert len(e_inst_list.instances) == 1, f'unsupported RTL coding style at line {e_inst_list.lineno}'
      e_node = e_inst_list.instances[0]
      self.inst_name_to_rtl[e_node.name] = self.codegen.visit(e_inst_list)
      self.e_name_to_ast_node[e_node.name] = e_inst_list

  # get mapping from reg/wire/io name to width/direction
  def __initDeclList(self):
    # replace the parameters in 
    def replaceParamByValue(width : str):
      tokens = re.findall(r'[A-Za-z0-9_]+', width)
      while any(t in self.param_to_value_str for t in tokens):
        for t in tokens:
          if t in self.param_to_value_str:
            width = width.replace(t, self.param_to_value_str[t] )
            logging.debug(f'replace {t} by {self.param_to_value_str[t]}')
        tokens = re.findall(r'[A-Za-z0-9_]+', width)
      return width.replace(' ', '') # must remove the spaces because later steps will partition expressions based on space

    # get a copy of all delcaration
    for decl_node in self.__DFS(self.top_module_ast, lambda node : isinstance(node, ast.Decl)):
      assert len(decl_node.children()) == 1
      content = decl_node.children()[0]
      name = content.name

      if not isinstance(content, ast.Input) and not isinstance(content, ast.Output):
        self.all_decl_except_io.append(self.codegen.visit(decl_node))

    # initialize various mappings
    for decl_node in self.__DFS(self.top_module_ast, lambda node : isinstance(node, ast.Decl)):
      assert len(decl_node.children()) == 1
      content = decl_node.children()[0]
      name = content.name

      # filter out ast.Parameter
      if isinstance(content, ast.Parameter):
        continue
      
      width_node = content.width
      if width_node:
        width = self.codegen.visit(width_node)
        width = replaceParamByValue(width)
      else:
        width = ''

      if isinstance(content, ast.Input):
        self.io_name_to_dir[name] = 'input'
        self.io_name_to_width[name] = width

      elif isinstance(content, ast.Output):
        self.io_name_to_dir[name] = 'output'
        self.io_name_to_width[name] = width

      elif isinstance(content, ast.Wire) or isinstance(content, ast.Reg):
        self.reg_wire_name_to_width[name] = width

      else:
        logging.debug(f'unrecorded Decl statement: {name} @ line {decl_node.lineno}')

  def __initParamToValueStr(self):
    for decl_node in self.__DFS(self.top_module_ast, lambda node : isinstance(node, ast.Decl)):
      assert len(decl_node.children()) == 1
      content = decl_node.children()[0]
      
      if isinstance(content, ast.Parameter):
        self.param_to_value_str[content.name] = self.codegen.visit(content.value)

  def __initFIFOListOfModuleInst(self):
    for v_node in self.traverseVertexInAST():
      for portarg in v_node.portlist:
        # filter out constant ports
        if(not isinstance(portarg.argname, ast.Identifier)):
          continue

        port_name = portarg.portname
        wire_name = portarg.argname.name
        
        # each fifo xxx -> xxx_din & xxx_dout, each maps to a vertex
        # note that 'dout' is the output side of FIFO, thus the input side for the vertex
        if '_dout' in port_name:
          assert '_dout' in wire_name
          fifo_name = self.wire_to_fifo_name[wire_name]
          self.mod_to_fifo_in[v_node.name].append(fifo_name) 

        elif '_din' in port_name:
          assert '_din' in wire_name
          fifo_name = self.wire_to_fifo_name[wire_name]
          self.mod_to_fifo_out[v_node.name].append(fifo_name) 
          
        else:
          continue

  def __initWireToFIFOMapping(self):
    outbound_tags = ['_read', '_dout', '_empty_n']
    inbound_tags = ['_write', '_din', '_full_n']
    ctrl_tags = ['clk', 'rst', 'reset']

    for e_node in self.traverseEdgeInAST():
      for portarg in e_node.portlist:
        # filter constant ports
        if(not isinstance(portarg.argname, ast.Identifier)):
          continue

        port_name = portarg.portname
        wire_name = portarg.argname.name
        self.wire_to_fifo_name[wire_name] = e_node.name
        self.fifo_name_to_wires[e_node.name].append((port_name, wire_name))

        if any(o_tag in port_name for o_tag in outbound_tags):
          self.fifo_name_to_outbound_wires[e_node.name].append(wire_name)
        elif any(i_tag in port_name for i_tag in inbound_tags):
          self.fifo_name_to_inbound_wires[e_node.name].append(wire_name)
        else:
          assert any(c_tag in port_name for c_tag in ctrl_tags)

  def __initWireToVertexMapping(self):
    for v_node in self.traverseVertexInAST():
      for portarg in v_node.portlist:
        # filter constant ports
        if(not isinstance(portarg.argname, ast.Identifier)):
          continue

        wire_name = portarg.argname.name
        if 'ap_clk' in wire_name or 'ap_rst' in wire_name:
          continue
        
        # if two vertices are connected by direct wire, they must be bundled together
        if wire_name in self.wire_to_v_name:
          self.grouping_constraints.add( (self.wire_to_v_name[wire_name], v_node.name) )
          logging.info(f'grouping {self.wire_to_v_name[wire_name]} and {v_node.name} based on wire {wire_name}')

        self.wire_to_v_name[wire_name] = v_node.name
        self.v_name_to_wires[v_node.name].append(wire_name)

  # only some modules will be used to check if the whole design has finished
  # we only collect the ap_done names of those modules
  def __initApDoneSources(self):
    f = open(self.top_rtl_path, 'r')
    for line in f:
      if re.search(r'assign[ ]*ap_sync_done', line):
        line = re.sub(r'assign[ ]*ap_sync_done[ ]*=[ ]*', '', line)
        line = re.sub(r'[() ;\n]', '', line)
        ap_dones = line.split('&'); assert len(ap_dones)
        self.ap_done_v_name_to_wire = {self.wire_to_v_name[ap_done] : ap_done for ap_done in ap_dones}
        return
    assert False, 'ap_done signal in unexpected format'

  # extract which ap_ready signals will be used for the final ap_ready signal
  # [FIXME] some situations cannot be handled yet (multiple levels of intermediate wires)
  def __initApReadySources(self):
    f = open(self.top_rtl_path, 'r')
    
    # case 1: assign ap_ready = xxx & xxx ...
    for line in f:
      if re.search(r'assign[ ]*ap_ready', line):
        line = re.sub(r'assign[ ]*ap_ready[ ]*=[ ]*', '', line)
        line = re.sub(r'[() ;\n]', '', line)
        ap_readies = line.split('&'); assert len(ap_readies)
        
        if len(ap_readies) == 1 and ap_readies[0] == 'ap_sync_ready':
          break

        self.ap_ready_v_name_to_wire = {self.wire_to_v_name[ap_ready] : ap_ready for ap_ready in ap_readies}
        return
    
    # case 2: assign ap_ready = ap_sync_ready;
    f.seek(0)
    for line in f:
      if re.search(r'assign[ ]*ap_sync_ready', line):
        line = re.sub(r'assign[ ]*ap_sync_ready[ ]*=[ ]*', '', line)
        line = re.sub(r'[() ;\n]', '', line)
        ap_readies = line.split('&'); assert len(ap_readies)
        
        try:
          self.ap_ready_v_name_to_wire = {self.wire_to_v_name[ap_ready] : ap_ready for ap_ready in ap_readies}
        except:
          logging.critical('failed in parsing ap_ready signals')
        return
    
    assert False, 'ap_ready signal in unexpected format'


  def __isVertexNode(self, node):
    return isinstance(node, ast.Instance) and 'fifo' not in node.module

  def __isEdgeNode(self, node):
    return isinstance(node, ast.Instance) and 'fifo' in node.module

  def __isVertexInstanceList(self, node):
    return isinstance(node, ast.InstanceList) and 'fifo' not in node.module
  
  def __isEdgeInstanceList(self, node):
    return isinstance(node, ast.InstanceList) and 'fifo' in node.module


  #                                                 #
  # ---------------- Public Methods --------------- #
  #                                                 #

  def traverseVertexInAST(self):
    yield from self.__DFS(self.top_module_ast, self.__isVertexNode)
  
  def traverseEdgeInAST(self):
    yield from self.__DFS(self.top_module_ast, self.__isEdgeNode)
  
  # get the interface wires of vertex or edges
  def getWiresOfFIFOName(self, inst_name) -> list:
    return self.fifo_name_to_wires[inst_name]

  def getOutboundSideWiresOfFIFOName(self, fifo_name):
    return self.fifo_name_to_outbound_wires[fifo_name]

  def getInboundSideWiresOfFIFOName(self, fifo_name):
    return self.fifo_name_to_inbound_wires[fifo_name]

  def getWiresOfVertexName(self, v_name) -> list:
    return self.v_name_to_wires[v_name]

  def getWidthOfRegOrWire(self, name):
    width = self.reg_wire_name_to_width[name]
    assert ' ' not in width, 'Spaces in width express will result in error in getSlotNameToIOList()'
    return width

  def getIntegerWidthOfRegOrWire(self, name):
    width = self.reg_wire_name_to_width[name]
    if not width:
      return 1
    else:
      match = re.search(r'\[(\d+):',width).group(1)
      assert match
      return int(match)+1

  def getRTLOfInst(self, inst_name : str):
    return self.inst_name_to_rtl[inst_name]

  # transform the default FIFO used by HLS to our unified template
  def getFIFOInstOfNewTemplate(self, e_name:str, e_width : int, e_depth : int, pipeline_level : int, fifo_type: str = ""):
    e_inst_list = self.e_name_to_ast_node[e_name]

    # use our FIFO template
    e_inst_list.module = 'fifo_almost_full'

    param_width = ast.ParamArg('DATA_WIDTH', ast.Rvalue(ast.IntConst(str(e_width))) )
    param_depth = ast.ParamArg('DEPTH', ast.Rvalue(ast.IntConst(str(e_depth))) )

    addr_bitwidth = int(math.log2(e_width)+1)
    param_addr_width = ast.ParamArg( 'ADDR_WIDTH', ast.Rvalue(ast.IntConst(str(addr_bitwidth))) )

    # pipeline_level equals the required grace period
    param_grace_period = ast.ParamArg( 'GRACE_PERIOD', ast.Rvalue(ast.IntConst(str( pipeline_level ))) )
    
    if fifo_type == 'BRAM':
      param_grace_period = ast.ParamArg( 'USE_BRAM', ast.Rvalue(ast.IntConst(str( 1 ))) )
    elif fifo_type == 'SRL':
      param_grace_period = ast.ParamArg( 'USE_SRL', ast.Rvalue(ast.IntConst(str( 1 ))) )

    params = [param_width, param_depth, param_addr_width, param_grace_period]
    e_inst_list.parameterlist = params

    for c in e_inst_list.instances: # should only has 1 instance
      c.module = e_inst_list.module
      c.parameterlist = params

    return self.codegen.visit(e_inst_list)

  def getWidthOfIO(self, io_name):
    return self.io_name_to_width[io_name]

  def getAllDeclExceptIO(self):
    return self.all_decl_except_io

  def isIO(self, io_name):
    return io_name in self.io_name_to_width

  def getDirOfIO(self, io_name):
    return self.io_name_to_dir[io_name]

  def getDirWidthNameOfAllIO(self):
    top_io = []
    for name, width in self.io_name_to_width.items():
      if width:
        top_io.append((self.io_name_to_dir[name], width, name))
      else:
        top_io.append((self.io_name_to_dir[name], name))
    return top_io
    
  def getInFIFOsOfModuleInst(self, inst_name):
    return self.mod_to_fifo_in[inst_name]

  def getOutFIFOsOfModuleInst(self, inst_name):
    return self.mod_to_fifo_out[inst_name]

  # fifo_w32_d2_A xxx -> 32
  def getFIFOWidthFromFIFOType(self, fifo_type):
    match = re.search(r'_w(\d+)_d(\d+)_', fifo_type)
    assert match, f'wrong FIFO instance name: {fifo_type}'
    return int(match.group(1)) # group 111111

  # fifo_w32_d2_A xxx -> 2
  def getFIFODepthFromFIFOType(self, fifo_type):
    match = re.search(r'_w(\d+)_d(\d+)_', fifo_type)
    assert match, f'wrong FIFO instance name: {fifo_type}'
    return int(match.group(2)) # group 222222

  def getFIFONameFromInstanceList(self, node):
    assert(len(node.instances) == 1)
    return node.instances[0].name

  # only contains the ap_done signals that are part of the final ap_done
  def getApDoneVNameToWire(self):
    return self.ap_done_v_name_to_wire

  # only contains the ap_ready signals that are part of the final ap_ready
  def getApReadyVNameToWire(self):
    return self.ap_ready_v_name_to_wire

  def getParamValueStr(self, param_name):
    return self.param_to_value_str[param_name]

  def getParamToValueStr(self):
    return self.param_to_value_str

  def getFIFONameFromWire(self, wire):
    return self.wire_to_fifo_name[wire]

  # if two vertices are connected by direct wires
  # they must always remain in the same slot
  def getStrictGroupingConstraints(self):
    return list(self.grouping_constraints)

  def getTopModuleName(self):
    return self.top_rtl_path.split('/')[-1]


import logging
from collections import defaultdict
from typing import Dict, List, Tuple
from mip import Model, minimize, BINARY, xsum, OptimizationStatus, Var

from autobridge.Opt.Common import RESOURCE_TYPES
from autobridge.Opt.DataflowGraph import Vertex
from autobridge.Opt.Slot import Slot

_logger = logging.getLogger().getChild(__name__)


def _createILPVars(
    m: Model, v_list: List[Vertex], s_list: List[Slot]
) -> Tuple[Dict[Vertex, Dict[Slot, Var]], Dict[Slot, Dict[Vertex, Var]]]:
  """
  for each vertex, for each slot, create a binary var if the vertex is assigned to the slot
  """
  _logger.info('Creating ILP variables...')
  v_to_s_to_var = defaultdict(dict)
  s_to_v_to_var = defaultdict(dict)
  for v in v_list:
    for s in s_list:
      v_to_s_to_var[v][s] = m.add_var(var_type=BINARY)
      s_to_v_to_var[s][v] = v_to_s_to_var[v][s]

  return v_to_s_to_var, s_to_v_to_var


def _addAreaConstrains(
    m: Model, 
    s_to_v_to_var: Dict[Slot, Dict[Vertex, Var]],
    resource_usage_limit: int
) -> None:
  """
  limit the capacity of each slot
  """
  _logger.info('Adding area constraints...')
  for r in RESOURCE_TYPES:
    for s, v_to_var in s_to_v_to_var.items():
      capacity = s.area[r] * resource_usage_limit
      m += xsum(v.area[r] * var for v, var in v_to_var.items()) <= capacity


def _addUniqueConstrains(m: Model, v_to_s_to_var: Dict[Vertex, Dict[Slot, Var]]) -> None:
  """
  each vertex is assigned to one slot
  """
  _logger.info('Adding constraints that each Vertex is assigned to one slot...')
  for v, s_to_var in v_to_s_to_var.items():
    m += xsum(var for var in s_to_var.values()) == 1


def _getVToSToCost(
    v_list: List[Vertex], s_list: List[Slot], orig_v2s: Dict[Vertex, Slot]
) -> Dict[Vertex, Dict[Slot, int]]:
  """
  cost for assigning a vertex to a slot
  Define the cost as the distance from the original location * the total wire length
  """
  _logger.info('Generating cost...')
  v_to_s_to_cost = defaultdict(dict)
  for v in v_list:
    for s in s_list:
      orig_s = orig_v2s[v]
      # +1 to handle the case where a module has 0 wire length
      v_to_s_to_cost[v][s] = orig_s.getDistance(s) * (v.getTotalWireWidth() + 1)

  return v_to_s_to_cost


def _addObjective(
    m: Model, 
    v_to_s_to_cost: Dict[Vertex, Dict[Slot, int]], 
    v_to_s_to_var: Dict[Vertex, Dict[Slot, Var]]
) -> None:
  """
  minimize the cost
  """
  _logger.info('Adding objective...')
  cost_var_pair_list: List[Tuple[int, Var]] = []
  for v, s_to_var in v_to_s_to_var.items():
    for s, var in s_to_var.items():
      cost = v_to_s_to_cost[v][s]
      cost_var_pair_list.append((cost, var))

  m.objective = minimize(xsum(cost * var for cost, var in cost_var_pair_list))


def _addGroupingConstraints(
    m: Model,
    grouping_list: List[List[Vertex]],
    v_to_s_to_var: Dict[Vertex, Dict[Slot, Var]],
    s_list: List[Slot]
) -> None:
  _logger.info('Add grouping constraints...')

  for grouping in grouping_list:
    for i in range(1, len(grouping)):
      v1 = grouping[0]
      v2 = grouping[i]
      _logger.info(f'Grouping {v1.name} and {v2.name}')
      for s in s_list:
        m += v_to_s_to_var[v1][s] == v_to_s_to_var[v2][s]


def _getILPResults(
    v_to_s_to_var: Dict[Vertex, Dict[Slot, int]]
) -> Tuple[ Dict[Vertex, Slot], Dict[Slot, List[Vertex]] ]:
  """
  extract which modules is assigned to which slots
  """
  _logger.info('Extracting ILP results...')
  # get v2s
  new_v2s = {}
  for v, s_to_var in v_to_s_to_var.items():
    for s, var in s_to_var.items():
      if round(var.x) == 1:
        new_v2s[v] = s
        break

    assert v in new_v2s

  # get s2v
  new_s2v = defaultdict(list)
  for v, s in new_v2s.items():
    new_s2v[s].append(v)

  return new_v2s, new_s2v


def _logResults(
    new_s2v: Dict[Slot, List[Vertex]], 
    new_v2s: Dict[Vertex, Slot], 
    orig_v2s: Dict[Vertex, Slot]
) -> None:
  """
  analyze and log the new floorplan results
  """
  for s, v_list in new_s2v.items():
    _logger.info(f'Slot {s.getRTLModuleName()}:')
    for r in RESOURCE_TYPES:
      capacity = s.area[r]
      usage = sum(v.area[r] for v in v_list)
      _logger.info(f'  [{r}]: {usage} / {capacity} = {round(usage/capacity, 5)} ')

  # log which vertices are re-placed
  for v, s in new_v2s.items():
    orig_s = orig_v2s[v]
    if s != orig_s:
      _logger.info(f'Vertex {v.name} is moved {orig_s.getDistance(s)} units from {orig_s.getRTLModuleName()} to {s.getRTLModuleName()}')


def legalizeFloorplanResults(
    orig_v2s: Dict[Vertex, Slot],
    grouping_list: List[ List[Vertex] ],
    all_slot_list: List[Slot],
    resource_usage_limit: int
) -> Tuple[ Dict[Slot, List[Vertex]], Dict[Vertex, Slot] ]:
  """
  adjust the floorplanning to satisfy the area requirement
  """
  _logger.info('Begin legalizing the floorplan results...')
  _logger.info(f'Target resource usage limit: {resource_usage_limit}')

  m = Model()

  v_list = list(orig_v2s.keys())
  s_list = all_slot_list

  v_to_s_to_var, s_to_v_to_var = _createILPVars(m, v_list, s_list)
  v_to_s_to_cost = _getVToSToCost(v_list, s_list, orig_v2s)

  _addAreaConstrains(m, s_to_v_to_var, resource_usage_limit)

  _addUniqueConstrains(m, v_to_s_to_var)

  _addGroupingConstraints(m, grouping_list, v_to_s_to_var, s_list)

  _addObjective(m, v_to_s_to_cost, v_to_s_to_var)

  status = m.optimize()
  if status != OptimizationStatus.OPTIMAL:
    _logger.warning(f'Fail to legalize the floorplan under target ratio {resource_usage_limit}')
    m.write('floorplan_legalization.lp')
    return None, None

  new_v2s, new_s2v = _getILPResults(v_to_s_to_var)

  _logResults(new_s2v, new_v2s, orig_v2s)

  _logger.info('Finish legalizing the floorplan results.')

  return new_s2v, new_v2s


def AutoLegalizer(
    orig_v2s: Dict[Vertex, Slot],
    grouping_list: List[ List[Vertex] ],
    all_slot_list: List[Slot],  # which slots are considered in the legalization step
    init_resource_usage_limit: float = 0.7,
    resource_usage_cut_threshold: float = 0.75,
    limit_increase_step: float = 0.01
) -> Tuple[ Dict[Slot, List[Vertex]], Dict[Vertex, Slot] ]:

  curr_limit = init_resource_usage_limit
  while 1:
    if curr_limit > resource_usage_cut_threshold:
      _logger.error(f'Fail to legalize under the cut threhold {resource_usage_cut_threshold}')
      exit()

    new_s2v, new_v2s = legalizeFloorplanResults(orig_v2s, grouping_list, all_slot_list, curr_limit)
    if new_s2v:
      _logger.info(f'Legalization succeeded with target usage limit {curr_limit}')
      return new_s2v, new_v2s
    else:
      curr_limit += limit_increase_step


# get the original top RTL
# locate the FIFO instance of all edges
# update the FIFO type to relay station
# set the pipeline level
from typing import Dict

import pyverilog.vparser.ast as ast
from autobridge.Opt.DataflowGraph import Edge


def _is_fifo_instance_list(node):
  return isinstance(node, ast.InstanceList) and 'fifo' in node.module


def _get_fifo_name_from_instance_list(node):
  assert(len(node.instances) == 1)
  return node.instances[0].name


def _convert_hls_fifo_to_relay_station(node, e: Edge) -> None:
  node.module = 'relay_station'  # pipeline the FIFO

  width = ast.ParamArg( 'DATA_WIDTH', ast.Rvalue(ast.IntConst(str(e.width))) )
  depth = ast.ParamArg( 'DEPTH', ast.Rvalue(ast.IntConst(str(e.depth + e.added_depth_for_rebalance))) )
  addr_width = ast.ParamArg( 'ADDR_WIDTH', ast.Rvalue(ast.IntConst(str(e.addr_width))) )
  level = ast.ParamArg( 'LEVEL', ast.Rvalue(ast.IntConst(str( e.pipeline_level ))) )

  params = [width, depth, addr_width, level]
  node.parameterlist = params

  for c in node.instances:
    c.module = 'relay_station'
    c.parameterlist = params


def _convert_hls_fifo_to_autobridge_fifo_template(node, e: Edge) -> None:
  # TODO: remove the extra grace period from the FIFO template

  node.module = 'fifo_almost_full'  # pipeline the FIFO

  width = ast.ParamArg( 'DATA_WIDTH', ast.Rvalue(ast.IntConst(str(e.width))) )
  depth = ast.ParamArg( 'DEPTH', ast.Rvalue(ast.IntConst(str(e.depth + e.added_depth_for_rebalance))) )
  addr_width = ast.ParamArg( 'ADDR_WIDTH', ast.Rvalue(ast.IntConst(str(e.addr_width))) )
  grace_period = ast.ParamArg( 'GRACE_PERIOD', ast.Rvalue(ast.IntConst(str( "0" ))) )

  params = [width, depth, addr_width, grace_period]
  node.parameterlist = params

  for c in node.instances:
    c.module = 'fifo_almost_full'
    c.parameterlist = params


def convert_fifo(node, edge_name_to_object: Dict[str, Edge]):
  # only considers fifo/rs instances
  if (not _is_fifo_instance_list(node)):
    return 

  edge_name = _get_fifo_name_from_instance_list(node)
  e = edge_name_to_object[edge_name]

  if e.pipeline_level == 0:
    _convert_hls_fifo_to_autobridge_fifo_template(node, e)
  else:
    _convert_hls_fifo_to_relay_station(node, e)


def add_keep_hierarcy_pragma(node):  #non fifo modules
  if (not isinstance(node, ast.InstanceList)):
    return

  # do not keep hierarchy of relay stations
  # because they are meant to be split and span long distances
  if node.module == 'relay_station':
    return
      
  node.module = f'(* keep_hierarchy = "yes" *) {node.module}'
  for c in node.instances:
    c.module = f'(* keep_hierarchy = "yes" *) {c.module}'






#!/usr/bin/python3
import collections
import json
import logging
import sys
from typing import List

import autobridge.Floorplan.Utilities as util
import autobridge.Floorplan as autobridge_floorplan
from autobridge.Device.DeviceManager import DeviceManager
from autobridge.HLSParser.tapa.DataflowGraphTapa import DataflowGraphTapa
from autobridge.HLSParser.tapa.ProgramJsonManager import ProgramJsonManager
from autobridge.Opt.Floorplan import Floorplanner
from autobridge.Opt.Slot import Topology
from autobridge.Opt.SlotManager import SlotManager

logging.basicConfig(
    level=logging.DEBUG,
    format=
    '%(levelname).1s%(asctime)s.%(msecs)03d %(name)s:%(lineno)d] %(message)s',
    datefmt='%m%d %H:%M:%S',
)

_logger = logging.getLogger().getChild(__name__)


def get_floorplan_output_bundle(config):
  assert config['CompiledBy'] == 'TAPA'

  board = DeviceManager(
      board_name=config['Board'],
      ddr_list=config.get('DDR', []),
      is_vitis_enabled=True,
  ).getBoard()
  program_json_manager = ProgramJsonManager(
      config['Edges'],
      config['Vertices'],
      config['Area'],
  )
  graph = DataflowGraphTapa(program_json_manager)
  slot_manager = SlotManager(board)

  # which modules must be assigned to the same slot
  grouping_constraints: List[List[str]] = config.get('GroupingConstraints', [])

  # process optional module pre-assignment constraints
  module_floorplan = config['OptionalFloorplan']
  pre_assignment = {}
  for region, module_group in module_floorplan.items():
    for mod_name in module_group:
      pre_assignment[mod_name] = region

  kwargs = {}
  ref_usage_ratio = config.get('MaxUsage')
  if ref_usage_ratio is not None:
    kwargs['ref_usage_ratio'] = ref_usage_ratio

  # generate floorplan
  v2s = autobridge_floorplan.get_floorplan(graph, slot_manager, grouping_constraints, pre_assignment, **kwargs)
  
  s2v = util.invert_v2s(v2s)
  topology = Topology(s2v)

  return v2s, s2v, topology

def generate_constraints(config):
  """ [obsolete] """
  v2s, s2v, topology = get_floorplan_output_bundle(config)
  return {
      slot.pblock_name:
      [v.name for v in v_group] + [topology.getTopologyOf(slot)]
      for slot, v_group in s2v.items()
  }

def generate_constraints_bundle(config):
  v2s, s2v, topology = get_floorplan_output_bundle(config)
  v_name_to_s_name = {v.name: s.pblock_name for v, s in v2s.items()}
  s_name_to_v_name = {s.pblock_name: [v.name for v in v_list] for s, v_list in s2v.items()}
  topology = {slot.pblock_name: topology.getTopologyOf(slot) for slot in s2v.keys()}
  return v_name_to_s_name, s_name_to_v_name, topology



def main():
  fp = sys.stdin
  if fp.isatty():
    fp = open('SampleUserConfig.json', 'r')
  constraints = generate_constraints(json.load(fp))

  fp = sys.stdout
  if fp.isatty():
    fp = open('TapaConstraint.json', 'w')
  json.dump(constraints, fp, indent=2)


if __name__ == '__main__':
  main()



import logging

from autobridge.Opt.DataflowGraph import Vertex, Edge
from autobridge.HLSParser.tapa.ProgramJsonManager import ProgramJsonManager

class DataflowGraphTapa ():

  def __init__(self, program_json_manager : ProgramJsonManager):
    self.program_json_manager = program_json_manager

    self.vertices = {} # name -> Vertex
    self.edges = {} # name -> Edge

    self.__initEdges()

    self.__initVertices()

    self.__linkEdgeAndVertex()
    
  def __initEdges(self):
    fifo_section = self.program_json_manager.getFIFOSection()

    # init Edge objects
    for e_name, info in fifo_section.items():
      e = Edge(e_name)
      e.width = info['width']
      e.depth = info['depth']
      self.edges[e_name] = e

  def __initVertices(self):
    v_name_to_module = self.program_json_manager.getVNameToModule()

    for v_name, v_module in v_name_to_module.items():
      v = Vertex(v_module, v_name)

      # get area
      v.area = self.program_json_manager.getAreaOfModule(v_module)
      
      fifo_section = self.program_json_manager.getFIFOSection()
      for e_name, info in fifo_section.items():
        if info['produced_by'] == v_name:
          v.out_edge_names.append(e_name)
        if info['consumed_by'] == v_name:
          v.in_edge_names.append(e_name)
          
      self.vertices[v_name] = v

  def __linkEdgeAndVertex(self):
    for v in self.vertices.values():
      for fifo_in_name in v.in_edge_names:
        fifo_in = self.edges[fifo_in_name]
        fifo_in.dst = v
        v.in_edges.append(fifo_in)
      for fifo_out_name in v.out_edge_names:
        fifo_out = self.edges[fifo_out_name]
        fifo_out.src = v
        v.out_edges.append(fifo_out)

  def printVertices(self):
    for v in self.vertices.values():
      logging.debug(f'{v.name}: {v.area}')
      for e in v.in_edges:
        logging.debug(f'  <- {e.name}')
      for e in v.out_edges:
        logging.debug(f'  -> {e.name}')

  def printEdges(self):
    for e in self.edges.values():
      logging.debug(f'{e.name}: {e.src.name} -> {e.dst.name}')

  def getAllVertices(self):
    return self.vertices.values()

  def getAllEdges(self):
    return self.edges.values()

  def getNameToVertexMap(self):
    return self.vertices

  def getNameToEdgeMap(self):
    return self.edges

  def getVertex(self, v_name):
    return self.vertices[v_name]


import logging

from typing import Dict, List, Callable, Optional, Tuple
from mip import Model, Var, minimize, xsum, BINARY, INTEGER, OptimizationStatus
from itertools import product

from autobridge.Floorplan.Utilities import *
from autobridge.Floorplan.ILPUtilities import *
from autobridge.Opt.DataflowGraph import Vertex
from autobridge.Opt.Slot import Slot
from autobridge.Opt.SlotManager import SlotManager, Dir

_logger = logging.getLogger().getChild(__name__)


def eight_way_partition(
  init_v2s: Dict[Vertex, Slot],
  slot_manager: SlotManager,
  grouping_constraints: List[List[Vertex]],
  pre_assignments: Dict[Vertex, Slot],
  ref_usage_ratio: float,
  max_search_time: int = 600,
  warm_start_assignments: Dict[Vertex, Slot] = {},
  max_usage_ratio: float = 0.84,
  ref_slr_width_limit: int = 10000,
  max_slr_width_limit: int = 15000,
) -> Optional[Dict[Vertex, Slot]]:
  """
  adjust the max_usage_ratio if failed
  """
  best_v2s = {}

  # start from the largest allowed ratio and search downwards. Prune if fail.
  for curr_max_usage in reversed(float_range(ref_usage_ratio, max_usage_ratio, 0.02)):
    _logger.info(f'Attempt eight way partition with max_usage_ratio {curr_max_usage}')

    curr_best_v2s = {}

    # start from the largest allowed slr crossing and search downwards.
    # if one iteration has solution, update the current solution and keep reducing the slr_limit
    # if one iteration does not have solution, then break and return the current solution
    for curr_slr_limit in reversed(range(ref_slr_width_limit, max_slr_width_limit, 500)):
      _logger.info(f'Try slr_width_limit {curr_slr_limit}')

      v2s = _eight_way_partition(
        init_v2s, 
        grouping_constraints, 
        pre_assignments, 
        slot_manager, 
        curr_max_usage, 
        max_search_time, 
        warm_start_assignments,
        curr_slr_limit,
      )

      if v2s:
        curr_best_v2s = v2s
      else:
        break

    if curr_best_v2s:
      _logger.info(f'Found solution with max_usage_ratio {curr_max_usage} and slr_width_limit {curr_slr_limit}')
      best_v2s = curr_best_v2s
    else:
      break

  if not best_v2s:
    _logger.info(f'eight way partition failed with max_usage_ratio {curr_max_usage} and slr_width_limit {max_slr_width_limit}')
    return {}

  log_resource_utilization(best_v2s)
  return best_v2s


def _eight_way_partition(
  init_v2s: Dict[Vertex, Slot],
  grouping_constraints: List[List[Vertex]],
  pre_assignments: Dict[Vertex, Slot],
  slot_manager: SlotManager,
  max_usage_ratio: float,
  max_search_time: int,
  warm_start_assignments: Dict[Vertex, Slot],
  slr_width_limit: int,
) -> Dict[Vertex, Slot]:

  m = Model()
  m.verbose = 0

  v_list = list(init_v2s.keys())

  # three variables could determine the location of a module
  # y = y1 *2 + y2  (four slots)
  # x = x           (each SLR is divided by half)
  v2var_x, v2var_y1, v2var_y2 = dict(), dict(), dict()
  for v in v_list:
    v2var_x[v] = m.add_var(var_type=BINARY, name=f'{v.name}_x')
    v2var_y1[v] = m.add_var(var_type=BINARY, name=f'{v.name}_y1')
    v2var_y2[v] = m.add_var(var_type=BINARY, name=f'{v.name}_y2')

  func_get_slot_by_idx = _get_slot_by_idx_closure(slot_manager)
  slot_to_idx = _get_slot_to_idx(func_get_slot_by_idx)

  _add_area_constraints(m, v_list, v2var_x=v2var_x, v2var_y1=v2var_y1, v2var_y2=v2var_y2,
    func_get_slot_by_idx=func_get_slot_by_idx, max_usage_ratio=max_usage_ratio)

  _add_pre_assignment(m, v_list, slot_to_idx, pre_assignments, v2var_x=v2var_x, v2var_y1=v2var_y1, v2var_y2=v2var_y2)

  add_slr_0_1_crossing_constraint(m, v_list, v2var_y1, v2var_y2, slr_width_limit)
  add_slr_1_2_crossing_constraint(m, v_list, v2var_y1, slr_width_limit)
  add_slr_2_3_crossing_constraint(m, v_list, v2var_y1, v2var_y2, slr_width_limit)

  _add_grouping_constraints(m, grouping_constraints, v2var_x=v2var_x, v2var_y1=v2var_y1, v2var_y2=v2var_y2)

  _add_opt_goal(m, v_list, v2var_x=v2var_x, v2var_y1=v2var_y1, v2var_y2=v2var_y2)

  if warm_start_assignments:
    _add_warm_start_assignment(m, v_list, slot_to_idx, warm_start_assignments, v2var_x=v2var_x, v2var_y1=v2var_y1, v2var_y2=v2var_y2)

  _logger.debug(f'Start ILP solver with max usage ratio {max_usage_ratio} and max search time {max_search_time}s')
  m.optimize(max_seconds=max_search_time)

  next_v2s = _get_results(m, v_list, func_get_slot_by_idx, v2var_x=v2var_x, v2var_y1=v2var_y1, v2var_y2=v2var_y2)

  return next_v2s


def _get_slot_by_idx_closure(
  slot_manager: SlotManager
) -> Callable[[int, int, int], Slot]:
  # slot_group = [slot_000, slot_001, \
  #               slot_010, slot_011, \
  #               slot_100, slot_101, \
  #               slot_110, slot_111 ]

  # must not change order!
  partition_order = [Dir.horizontal, Dir.horizontal, Dir.vertical]
  all_leaf_slots = slot_manager.getLeafSlotsAfterPartition(partition_order)
  slr_slots = slot_manager.getLeafSlotsAfterPartition([Dir.horizontal, Dir.horizontal])

  # include both whole-slr slots and half-slr slots
  def func_get_slot_by_idx(y1, y2, x):
    if x != -1:
      idx = y1 * 4 + y2 * 2 + x
      return all_leaf_slots[idx]
    else:
      idx = y1 * 2 + y2
      return slr_slots[idx]

  return func_get_slot_by_idx


def _get_slot_to_idx(
  func_get_slot_by_idx: Callable[[int, int, int], Slot],
) -> Dict[Slot, Tuple[int, int, int]]:
  """
  given a slot, get (y2, y1, x) in a tuple
  if x == -1, the slot is a whole-slr slot
  """
  slot_to_idx = {}
  for y1, y2, x in product(range(2), range(2), range(2)):
    slot_to_idx[func_get_slot_by_idx(y1, y2, x)] = (y1, y2, x)

  for y1, y2 in product(range(2), range(2)):
    slot_to_idx[func_get_slot_by_idx(y1, y2, -1)] = (y1, y2, -1)

  return slot_to_idx


def _add_area_constraints(
  m: Model,
  v_list: List[Vertex],
  v2var_x: Dict[Vertex, Var],
  v2var_y1: Dict[Vertex, Var],
  v2var_y2: Dict[Vertex, Var],
  func_get_slot_by_idx: Callable[[int, int, int], Slot],
  max_usage_ratio: float,
) -> None:

  # area constraint
  for r in RESOURCE_TYPES:
    choose = lambda x, num: x if num == 1 else (1-x)

    for y1, y2, x in product(range(2), range(2), range(2)):
      # convert logic AND to linear constraints
      # prods[v] = choose_y1 AND choose_y2 AND choose_x
      prods = { v : m.add_var(var_type=BINARY, name=f'{v.name}_choose{y1}{y2}{x}') for v in v_list }
      for v in v_list:
        m +=  choose(v2var_y1[v], y1) + choose(v2var_y2[v], y2) + \
              choose(v2var_x[v], x) - 3 * prods[v] >= 0
        m +=  choose(v2var_y1[v], y1) + choose(v2var_y2[v], y2) + \
              choose(v2var_x[v], x) - 3 * prods[v] <= 2

      m += xsum(  prods[v] * v.getVertexAndInboundFIFOArea()[r] for v in v_list ) \
                  <= func_get_slot_by_idx(y1, y2, x).getArea()[r] * max_usage_ratio


def _add_grouping_constraints(
  m: Model,
  grouping_constraints: List[List[Vertex]],
  v2var_x: Dict[Vertex, Var],
  v2var_y1: Dict[Vertex, Var],
  v2var_y2: Dict[Vertex, Var],
) -> None:
  """
  user specifies that certain Vertices must be assigned to the same slot
  """
  for grouping in grouping_constraints:
    for i in range(1, len(grouping)):
      for v2var in [v2var_x, v2var_y1, v2var_y2]:
        m += v2var[grouping[0]] == v2var[grouping[i]]


def _add_pre_assignment(
  m: Model,
  v_list: List[Vertex],
  slot_to_idx: Dict[Slot, Tuple[int, int, int]],
  pre_assignments: Dict[Vertex, Slot],
  v2var_x: Dict[Vertex, Var],
  v2var_y1: Dict[Vertex, Var],
  v2var_y2: Dict[Vertex, Var],
) -> None:
  # check if the pre assignment is correct
  v_set = set(v_list)
  for v, expect_slot in pre_assignments.items():
    assert v in v_set, f'ERROR: user has forced the location of a non-existing module {v.name}'

  # consider two types of pre-assignment
  # if x == 0 or 1, that means a half-SLR slot
  # if x == -1, that means a whole-SLR slot
  for v, expect_slot in pre_assignments.items():
    y1, y2, x = slot_to_idx[expect_slot]
    m += v2var_y1[v] == y1
    m += v2var_y2[v] == y2

    # only add constraints for x if the slot is half-slr
    if x != -1:
      m += v2var_x[v] == x


def _add_warm_start_assignment(
  m: Model,
  v_list: List[Vertex],
  slot_to_idx: Dict[Slot, Tuple[int, int, int]],
  warm_start_assignment: Dict[Vertex, Slot],
  v2var_x: Dict[Vertex, Var],
  v2var_y1: Dict[Vertex, Var],
  v2var_y2: Dict[Vertex, Var],
) -> None:
  """
  provide an existing solution as a warm start to the ILP solver
  WARNING: not useful at the moment
  """
  assert all(v in warm_start_assignment for v in v_list), 'ERROR: incomplete initial solution'

  warm_start: List[Tuple[Var, int]] = []
  for v, init_sol_slot in warm_start_assignment.items():
    y1, y2, x = slot_to_idx[init_sol_slot]
    warm_start += [(v2var_y1[v], y1), (v2var_y2[v], y2), (v2var_x[v], x)]
  m.start = warm_start


def _add_opt_goal(
  m: Model,
  v_list: List[Vertex],
  v2var_x: Dict[Vertex, Var],
  v2var_y1: Dict[Vertex, Var],
  v2var_y2: Dict[Vertex, Var],
) -> None:
  # add optimization goal
  all_edges = get_all_edges(v_list)
  e2cost_var = {e: m.add_var(var_type=INTEGER, name=f'intra_{e.name}') for e in all_edges}

  # note pos is different from slot_idx, becasue the x dimension is different from the y dimention
  # we will use |(y1 * 2 + y1) - (y2 * 2 + y2)| + |x1 - x2| to express the hamming distance
  pos_y = lambda v : v2var_y1[v] * 2 + v2var_y2[v]
  pos_x = lambda v : v2var_x[v]
  cost_y = lambda e : pos_y(e.src) - pos_y(e.dst)
  cost_x = lambda e : pos_x(e.src) - pos_x(e.dst)

  for e, cost_var in e2cost_var.items():
    m += cost_var >= cost_y(e) + cost_x(e)
    m += cost_var >= -cost_y(e) + cost_x(e)
    m += cost_var >= cost_y(e) - cost_x(e)
    m += cost_var >= -cost_y(e) - cost_x(e)

  m.objective = minimize(xsum(cost_var * e.width for e, cost_var in e2cost_var.items() ) )


def _get_results(
  m: Model,
  v_list: List[Vertex],
  func_get_slot_by_idx: Callable[[int, int, int], Slot],
  v2var_x: Dict[Vertex, Var],
  v2var_y1: Dict[Vertex, Var],
  v2var_y2: Dict[Vertex, Var],
) -> Optional[Dict[Vertex, Slot]]:
  if m.status == OptimizationStatus.OPTIMAL:
    _logger.info(f'succeed with optimal solution')
  elif m.status == OptimizationStatus.FEASIBLE:
    _logger.info(f'finish with non-optimal solution')
  else:
    _logger.info(f'failed')
    return {}

  # extract results
  next_v2s = {}
  for v in v_list:
    selected_slot = func_get_slot_by_idx(int(v2var_y1[v].x), int(v2var_y2[v].x), int(v2var_x[v].x))
    next_v2s[v] = selected_slot

  return next_v2s


"""AutoBridge"""

from setuptools import find_packages, setup

# with open('../README.md', encoding='utf-8') as f:
#   long_description = f.read()

setup(
    name='autobridge',
    version='0.0.20220314.dev',
    description='AutoBridge',
    long_description='',
    long_description_content_type='text/markdown',
    url='https://github.com/Licheng-Guo/AutoBridge',
    author='Licheng Guo',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: System :: Hardware',
    ],
    packages=find_packages(),
    python_requires='>=3.6',
    install_requires=[
        'mip',
        'pyverilog',
    ],
    entry_points={
        'console_scripts': [],
    },
)



from typing import Dict, List

from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
from autobridge.Codegen.PipeliningFIFO import convert_fifo, add_keep_hierarcy_pragma
from autobridge.Codegen.PipeliningControl import (
  pipeline_ap_rst, 
  pipeline_ap_start, 
  pipeline_ap_done,
  collect_all_ap_done_signals,
  make_ap_ready_equal_to_ap_done,
  make_ap_idle_equal_to_ap_done,
  fix_ap_continue_to_ap_done,
)
from autobridge.Codegen.InjectPipelineLogic import (
  get_ap_start_pipeline_def,
  get_ap_rst_pipeline_def,
  get_ap_done_pipeline_def,
  remove_orig_ctrl_signal,
)
from autobridge.Codegen.FIFOTemplate import fifo_template
from autobridge.Opt.DataflowGraph import Edge


def get_rtl(top_mod_ast) -> List[str]:
  codegen = ASTCodeGenerator()
  result = codegen.visit(top_mod_ast)

  # need the timescale command for simulation
  return ['`timescale 1 ns / 1 ps \n'] + result.split('\n')


def level_traverse(node, func, *arg):
  q = [node]
  while( len(q) ):
    curr = q.pop(0)
    for c in curr.children():
      q.append(c)
    func(curr, *arg)


def inject_rtl(curr_rtl: List[str], new_rtl: List[str]) -> None:
  """ inject the new RTL immediately after IO definition """
  for i in range(1, len(curr_rtl)):
    if not curr_rtl[i].strip():
      if 'reg ' in curr_rtl[i-1] or 'wire ' in curr_rtl[i-1]:
        curr_rtl[i+1:i+1] = new_rtl
        return
  assert False


def generate_new_rtl_top(top_mod_ast, edge_name_to_object: Dict[str, Edge], output_filename: str, floorplan):
  v_name_to_s = floorplan.getVertexNameToSlot()
  e_name_to_s = floorplan.getEdgeNameToSlot()
  slot_list = list(floorplan.s2v.keys())

  # first edit the signals connected to each module instance
  level_traverse(top_mod_ast, convert_fifo, edge_name_to_object)
  level_traverse(top_mod_ast, add_keep_hierarcy_pragma)
  level_traverse(top_mod_ast, pipeline_ap_rst, v_name_to_s, e_name_to_s)
  level_traverse(top_mod_ast, pipeline_ap_start, v_name_to_s)
  level_traverse(top_mod_ast, pipeline_ap_done)
  level_traverse(top_mod_ast, make_ap_ready_equal_to_ap_done)
  level_traverse(top_mod_ast, make_ap_idle_equal_to_ap_done)
  level_traverse(top_mod_ast, fix_ap_continue_to_ap_done)

  temp_rtl_top = get_rtl(top_mod_ast)

  # remove the previous assignment of ap_done and ap_ready
  remove_orig_ctrl_signal(temp_rtl_top)

  # then inject the pipeline logic
  pipeline_def = []
  pipeline_def += get_ap_start_pipeline_def(slot_list)
  pipeline_def += get_ap_rst_pipeline_def(slot_list)

  ap_done_module_list = []
  level_traverse(top_mod_ast, collect_all_ap_done_signals, ap_done_module_list)
  pipeline_def += get_ap_done_pipeline_def(v_name_to_s, ap_done_module_list)

  inject_rtl(temp_rtl_top, pipeline_def)

  # the relay station template
  temp_rtl_top.append(fifo_template)

  open(output_filename, 'w').write('\n'.join(temp_rtl_top))


#! /usr/bin/python3.6

import pyverilog.vparser.ast as ast

from typing import List


def pipeline_ap_rst(node, v_name_to_s, e_name_to_s):
  """
  update the signal connected to the module instances
  """
  if (not isinstance(node, ast.InstanceList)):
    return

  mod_name = node.instances[0].name
  curr_port_list = node.instances[0].portlist
  for portarg in curr_port_list:
    if ('ap_rst' in portarg.portname or 'reset' in portarg.portname or 'RESET' in portarg.portname):
      if 'relay_station' in node.instances[0].module or 'fifo' in node.instances[0].module:
        slot = e_name_to_s[mod_name]
      else:
        slot = v_name_to_s[mod_name]

      assert 'ap_rst_n' not in portarg.portname
      portarg.argname.name = f'ap_rst_{slot.getRTLModuleName()}'


def get_ap_rst_pipeline_rtl(slot_list, pipeline_level = 4) -> List[str]:
  """
  instantiate the pipeline registers
  """
  rtl = []
  for slot in slot_list:
    for i in range(pipeline_level):
      rtl.append(f'(* keep = "true" *) reg ap_rst_n_inv_{slot.getRTLModuleName()}_q{i};')
    
    rtl.append(f'always @ (posedge ap_clk) begin')
    rtl.append(f'  ap_rst_n_inv_{slot.getRTLModuleName()}_q0 <= ~ap_rst_n;')
    for i in range(1, pipeline_level):
      rtl.append(f'  ap_rst_n_inv_{slot.getRTLModuleName()}_q{i} <= ap_rst_n_inv_{slot.getRTLModuleName()}_q{i-1};')
    rtl.append(f'end')

    rtl.append(f'wire ap_rst_n_inv_{slot.getRTLModuleName()} = ap_rst_n_inv_{slot.getRTLModuleName()}_q{pipeline_level-1};')

  return rtl


def pipeline_ap_start(node, v_name_to_s):
  if (not isinstance(node, ast.InstanceList)):
    return

  mod_name = node.instances[0].name
  curr_port_list = node.instances[0].portlist
  for portarg in curr_port_list:
    if portarg.portname == 'ap_start':
      # handle the original ap_start from the s_axi_control
      if portarg.argname.name == 'ap_start':
        assert 's_axi' in mod_name
        portarg.argname.name = 'ap_start_orig'
        continue

      slot = v_name_to_s[mod_name]
      portarg.argname.name = f'ap_start_{slot.getRTLModuleName()}'


def pipeline_ap_done(node):
  if (not isinstance(node, ast.InstanceList)):
    return

  mod_name = node.instances[0].name
  curr_port_list = node.instances[0].portlist
  for portarg in curr_port_list:
    if portarg.portname == 'ap_done':
      # handle the original ap_start from the s_axi_control
      if portarg.argname.name == 'ap_done':
        assert 's_axi' in mod_name
        portarg.argname.name = 'ap_done_final'
        continue

      portarg.argname.name = f'ap_done_{mod_name}'
      return


def collect_all_ap_done_signals(node, ap_done_module_list: List[str]):
  if (not isinstance(node, ast.InstanceList)):
    return

  mod_name = node.instances[0].name
  curr_port_list = node.instances[0].portlist
  for portarg in curr_port_list:
    if portarg.portname == 'ap_done':
      # handle the original ap_start from the s_axi_control
      if 's_axi' in mod_name:
        continue

      ap_done_module_list.append(mod_name)
      return


def make_ap_ready_equal_to_ap_done(node):
  if (not isinstance(node, ast.InstanceList)):
    return

  curr_port_list = node.instances[0].portlist
  for portarg in curr_port_list:
    if portarg.portname == 'ap_ready':
      if portarg.argname.name == 'ap_ready':
        portarg.argname.name = 'ap_done_final'
      else:
        portarg.argname.name = ''


def make_ap_idle_equal_to_ap_done(node):
  if (not isinstance(node, ast.InstanceList)):
    return

  curr_port_list = node.instances[0].portlist
  for portarg in curr_port_list:
    if portarg.portname == 'ap_idle':
      if portarg.argname.name == 'ap_idle':
        portarg.argname.name = 'ap_done_final'
      else:
        portarg.argname.name = ''


def fix_ap_continue_to_ap_done(node):
  if (not isinstance(node, ast.InstanceList)):
    return

  curr_port_list = node.instances[0].portlist
  mod_name = node.instances[0].name

  for portarg in curr_port_list:
    if portarg.portname == 'ap_continue':
      portarg.argname.name = f'ap_done_{mod_name}'